# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)

"""
Data sets
---------

Defining data sets
^^^^^^^^^^^^^^^^^^

.. autoclass:: guidata.dataset.DataSet
    :members:

.. autoclass:: guidata.dataset.DataSetGroup
    :members:

.. autoclass:: guidata.dataset.ActivableDataSet
    :members:

Grouping items
^^^^^^^^^^^^^^

.. autoclass:: guidata.dataset.BeginGroup
    :members:

.. autoclass:: guidata.dataset.EndGroup
    :members:

.. autoclass:: guidata.dataset.BeginTabGroup
    :members:

.. autoclass:: guidata.dataset.EndTabGroup
    :members:

.. autoclass:: guidata.dataset.SeparatorItem
    :members:

Handling item properties
^^^^^^^^^^^^^^^^^^^^^^^^

.. autoclass:: guidata.dataset.ItemProperty

.. autoclass:: guidata.dataset.FormatProp
    :members:

.. autoclass:: guidata.dataset.GetAttrProp
    :members:

.. autoclass:: guidata.dataset.ValueProp
    :members:

.. autoclass:: guidata.dataset.NotProp
    :members:

.. autoclass:: guidata.dataset.FuncProp
    :members:

.. autoclass:: guidata.dataset.FuncPropMulti
    :members:
"""

# pylint: disable-msg=W0622
# pylint: disable-msg=W0212

from __future__ import annotations

import re
import sys
import warnings
from abc import ABC, abstractmethod
from collections.abc import Callable
from copy import deepcopy
from typing import TYPE_CHECKING, Any, TypeVar

import numpy as np

from guidata.config import ValidationMode, get_validation_mode
from guidata.io import INIReader, INIWriter
from guidata.userconfig import UserConfig

DEBUG_DESERIALIZE = False

if TYPE_CHECKING:
    from qtpy.QtCore import QSize
    from qtpy.QtWidgets import QDialog, QWidget

    from guidata.io import HDF5Reader, HDF5Writer, JSONReader, JSONWriter


class DataItemValidationWarning(UserWarning):
    """Warning raised when a DataItem value is invalid
    and validation mode is 'ENABLED'"""

    pass


class DataItemValidationError(ValueError):
    """Error raised when a DataItem value is invalid during validation.
    Provides more context about where the error occurred.

    Args:
        instance: the DataSet instance where the error occurred
        item: the DataItem that caused the error
        msg: error message
        operation: operation that failed, if applicable
    """

    def __init__(
        self, instance: DataSet, item: DataItem, msg: str, operation: str | None = None
    ) -> None:
        self.item = item
        self.operation = operation
        if operation:
            full_msg = f"{operation} failed for {item}: {msg}"
        else:
            full_msg = f"Checking {instance.__class__.__name__}.{item}: {msg}"
        super().__init__(full_msg)


class NoDefault:
    """Special value used to indicate that no default value is set for a DataItem"""

    pass


class ItemProperty:
    """Base class for item properties

    Args:
        callable (Callable): callable to use to evaluate the value of the property
    """

    def __init__(self, callable: Callable) -> None:
        self.callable = callable

    def __call__(self, instance: DataSet, item: Any, value: Any) -> Any:
        """Evaluate the value of the property given, the instance,
        the item and the value maintained in the instance by the item"""
        return self.callable(instance, item, value)

    def set(self, instance: DataSet, item: Any, value: Any) -> Any:
        """Sets the value of the property given an instance, item and value
        Depending on implementation the value will be stored either on the
        instance, item or self

        Args:
            instance (DataSet): instance of the DataSet
            item (Any): item to set the value of
            value (Any): value to set
        """
        raise NotImplementedError


FMT_GROUPS = re.compile(r"(?<!%)%\((\w+)\)")


class FormatProp(ItemProperty):
    """A Property that returns a string to help
    custom read-only representation of items

    Args:
        fmt (str): format string
        ignore_error (bool): ignore errors when formatting. Defaults to True.
    """

    def __init__(self, fmt: str, ignore_error: bool | None = True) -> None:
        """`fmt` is a format string
        it can contain a single anonymous substition or
        several named substitions.
        """
        self.fmt = fmt
        self.ignore_error = ignore_error
        self.attrs = FMT_GROUPS.findall(fmt)

    def __call__(self, instance: DataSet, item: DataItem, value: Any) -> Any:
        if not self.attrs:
            return self.fmt.format(value)
        dic = {}
        for attr in self.attrs:
            dic[attr] = getattr(instance, attr)
        try:
            return self.fmt % dic
        except TypeError:
            if not self.ignore_error:
                print(f"Wrong Format for {item._name} : {self.fmt!r} % {dic!r}")
                raise


class GetAttrProp(ItemProperty):
    """A property that matches the value of
    an instance's attribute

    Args:
        attr (str): attribute to match
    """

    def __init__(self, attr: str) -> None:
        self.attr = attr

    def __call__(self, instance: DataSet, item: DataItem, value: Any) -> Any:
        val = getattr(instance, self.attr)
        return val

    def set(self, instance: DataSet, item: DataItem, value: Any) -> None:
        setattr(instance, self.attr, value)


class ValueProp(ItemProperty):
    """A property that retrieves a value stored elsewhere

    Args:
        value (Any): value to store
    """

    def __init__(self, value: Any) -> None:
        self.value = value

    def __call__(self, instance: DataSet, item: DataItem, value: Any) -> Any:
        return self.value

    def set(self, instance: DataSet, item: DataItem, value: Any) -> None:
        """Sets the value of the property given an instance, item and value

        Args:
            instance (DataSet): instance of the DataSet
            item (Any): item to set the value of
            value (Any): value to set
        """
        self.value = value


class NotProp(ItemProperty):
    """Not property

    Args:
        prop (ItemProperty): property to negate
    """

    def __init__(self, prop: ItemProperty):
        self.property = prop

    def __call__(self, instance: DataSet, item: DataItem, value: Any) -> Any:
        return not self.property(instance, item, value)

    def set(self, instance: DataSet, item: DataItem, value: Any) -> None:
        """Sets the value of the property given an instance, item and value

        Args:
            instance (DataSet): instance of the DataSet
            item (Any): item to set the value of
            value (Any): value to set
        """
        self.property.set(instance, item, not value)


class FuncProp(ItemProperty):
    """An 'operator property'

    Args:
        prop (ItemProperty): property to apply function to
        func (function): function to apply
        invfunc (function): inverse function (default: func)
    """

    def __init__(
        self,
        prop: ItemProperty,
        func: Callable,
        invfunc: Callable | None = None,
    ) -> None:
        self.property = prop
        self.function = func
        if invfunc is None:
            invfunc = func
        self.inverse_function = invfunc

    def __call__(self, instance: DataSet, item: DataItem, value: Any) -> Any:
        return self.function(self.property(instance, item, value))

    def set(self, instance: DataSet, item: DataItem, value: Any) -> None:
        """Sets the value of the property given an instance, item and value

        Args:
            instance (DataSet): instance of the DataSet
            item (Any): item to set the value of
            value (Any): value to set
        """
        self.property.set(instance, item, self.inverse_function(value))


class FuncPropMulti(ItemProperty):
    """An 'operator property' for multiple properties

    Args:
        props (list[ItemProperty]): properties to apply function to
        func (function): function to apply
        invfunc (function): inverse function (default: func)
    """

    def __init__(
        self,
        props: list[ItemProperty],
        func: Callable,
        invfunc: Callable | None = None,
    ) -> None:
        self.properties = props
        self.function = func
        if invfunc is None:
            invfunc = func
        self.inverse_function = invfunc

    def __call__(self, instance: DataSet, item: DataItem, value: Any) -> Any:
        return self.function(*[prop(instance, item, value) for prop in self.properties])

    def set(self, instance: DataSet, item: DataItem, value: Any) -> None:
        """Sets the value of the property given an instance, item and value

        Args:
            instance (DataSet): instance of the DataSet
            item (Any): item to set the value of
            value (Any): value to set
        """
        for prop in self.properties:
            prop.set(instance, item, self.inverse_function(value))


class ComputedProp(ItemProperty):
    """A computed property that calls a method of the dataset to calculate values

    Args:
        method_or_name: name of the method to call on the dataset instance,
         or the function object itself
    """

    def __init__(self, method_or_name: str | callable) -> None:
        if callable(method_or_name):
            self.method = method_or_name
            self.method_name = getattr(method_or_name, "__name__", str(method_or_name))
        else:
            self.method = None
            self.method_name = method_or_name

    # pylint: disable=unused-argument
    def __call__(self, instance: DataSet, item: DataItem, value: Any) -> Any:
        """Compute the value by calling the specified method on the dataset instance

        Args:
            instance (DataSet): dataset instance
            item (DataItem): the data item
            value (Any): current value (ignored for computed items)

        Returns:
            Any: computed value
        """
        if self.method is not None:
            # Function object was provided directly
            return self.method(instance)
        else:
            # Method name was provided, get it from the instance
            method = getattr(instance, self.method_name, None)
            if method is None:
                raise AttributeError(
                    f"Dataset {instance.__class__.__name__} has no method "
                    f"'{self.method_name}'"
                )
            if not callable(method):
                raise TypeError(
                    f"Attribute '{self.method_name}' of {instance.__class__.__name__} "
                    "is not callable"
                )
            return method()

    # pylint: disable=unused-argument
    def set(self, instance: DataSet, item: DataItem, value: Any) -> None:
        """Computed properties cannot be set directly - they are read-only

        Args:
            instance (DataSet): instance of the DataSet
            item (DataItem): item to set the value of
            value (Any): value to set

        Raises:
            ValueError: Always, since computed items are read-only
        """
        raise ValueError(f"Computed item '{item.get_name()}' is read-only")


class DataItem(ABC):
    """DataSet data item

    Args:
        label (str): item label
        default (Any): default value
        help (str): text displayed on data item's tooltip
        check (bool): check value (default: True)
        allow_none (bool): if True, None values are allowed regardless of the
            expected type (default: False)
    """

    type = type
    count = 0

    def __init__(
        self,
        label: str,
        default: Any | None = None,
        help: str | None = "",
        check: bool | None = True,
        allow_none: bool = False,
    ) -> None:
        self._order = DataItem.count
        DataItem.count += 1
        self._name: str | None = None
        self._default = default
        self._help = help
        self._props: dict[
            Any, Any
        ] = {}  # a dict realm->dict containing realm-specific properties
        self.set_prop("display", col=0, colspan=None, row=None, label=label)
        self.set_prop("data", check_value=check, allow_none=allow_none)

    def get_prop(self, realm: str, name: str, default: Any = NoDefault) -> Any:
        """Get one property of this item

        Args:
            realm (str): realm name
            name (str): property name
            default (Any): default value (default: NoDefault)

        Returns:
            Any: property value
        """
        realm_props = self._props.get(realm)
        if realm_props is None:
            if default is NoDefault:
                raise KeyError(name)
            return default

        if default is NoDefault:
            return realm_props[name]  # Let KeyError propagate
        return realm_props.get(name, default)

    def get_prop_value(
        self, realm: str, instance: DataSet, name: str, default: Any = NoDefault
    ) -> Any:
        """Get one property of this item

        Args:
            realm (str): realm name
            instance (DataSet): instance of the DataSet
            name (str): property name
            default (Any): default value (default: NoDefault)

        Returns:
            Any: property value
        """
        value = self.get_prop(realm, name, default)
        if isinstance(value, ItemProperty):
            return value(instance, self, self.get_value(instance))
        else:
            return value

    def set_prop(self, realm: str, **kwargs) -> DataItem:
        """Set one or several properties using the syntax::

            set_prop(name1=value1, ..., nameX=valueX)

        It returns self so that we can assign to the result like this::

            item = Item().set_prop(x=y)

        Args:
            realm (str): realm name
            kwargs: properties to set

        Returns:
            DataItem: self
        """  # noqa
        prop = self._props.setdefault(realm, {})
        prop.update(kwargs)
        return self

    def set_pos(
        self, col: int = 0, colspan: int | None = None, row: int | None = None
    ) -> DataItem:
        """Set data item's position on a GUI layout

        Args:
            col (int): column number (default: 0)
            colspan (int): number of columns (default: None)
            row (int): row number (default: None)
        """
        self.set_prop("display", col=col, colspan=colspan, row=row)
        return self

    def set_computed(self, method_or_name: str | callable) -> DataItem:
        """Set data item as computed using the specified method

        Args:
            method_or_name: name of the method to call on the dataset instance to
             compute the value, or the function object itself

        Returns:
            DataItem: self
        """
        computed_prop = ComputedProp(method_or_name)
        self.set_prop("data", computed=computed_prop)
        # Also make it readonly in the display
        self.set_prop("display", readonly=True)
        return self

    def __str__(self) -> str:
        return f"{self._name} : {self.__class__.__name__}"

    def get_help(self, instance: DataSet) -> str:
        """Return data item's tooltip

        Args:
            instance (DataSet): instance of the DataSet

        Returns:
            str: tooltip
        """
        auto_help = self.get_auto_help(instance)
        help = self._help or ""
        if auto_help:
            help = help + "\n(" + auto_help + ")" if help else auto_help.capitalize()
        return help

    def get_auto_help(self, instance: DataSet) -> str:
        """Return the automatically generated part of data item's tooltip

        Args:
            instance (DataSet): instance of the DataSet

        Returns:
            str: automatically generated part of tooltip
        """
        return ""

    def format_string(self, instance: Any, value: Any, fmt: str, func: Callable) -> str:
        """Apply format to string representation of the item's value

        Args:
            instance (Any): instance of the DataSet
            value (Any): item's value
            fmt (str): format string
            func (Callable): function to apply to the value before formatting

        Returns:
            str: formatted string
        """
        return fmt % (func(value),)

    def get_string_value(self, instance: DataSet) -> str:
        """Return a formatted unicode representation of the item's value
        obeying 'display' or 'repr' properties

        Args:
            instance (DataSet): instance of the DataSet

        Returns:
            str: formatted string
        """
        value = self.get_value(instance)
        repval = self.get_prop_value("display", instance, "repr", None)
        if repval is not None:
            return repval
        else:
            fmt = self.get_prop_value("display", instance, "format", "%s")
            fmt = "%s" if fmt is None else fmt
            func = self.get_prop_value("display", instance, "func", lambda x: x)
            if (
                isinstance(fmt, Callable)  # type:ignore
                and value is not None
            ):
                return fmt(func(value))
            if value is not None:
                text = self.format_string(instance, value, fmt, func)
            else:
                text = "-"
            return text

    def get_name(self) -> str:
        """Return data item's name

        Returns:
            str: name
        """
        return self._name or ""

    def set_name(self, new_name: str) -> None:
        """Set data item's name

        Args:
            new_name (str): new name
        """
        self._name = new_name

    def set_help(self, new_help: str) -> None:
        """Set data item's help text

        Args:
            new_help (str): new help text
        """
        self._help = new_help

    def set_from_string(self, instance: DataSet, string_value: str) -> None:
        """Set data item's value from specified string

        Args:
            instance (DataSet): instance of the DataSet
            string_value (str): string value
        """
        value = self.from_string(string_value)
        self.__set__(instance, value)

    def get_default(self) -> Any:
        """Return data item's default value

        Returns:
            Any: default value
        """
        return self._default

    def set_default(self, instance: DataSet) -> None:
        """Set data item's value to default

        Args:
            instance (DataSet): instance of the DataSet
        """
        try:
            value = deepcopy(self._default)
            self._set_value_with_validation(instance, value, force_allow_none=True)
        except ValueError as exc:
            # Convert generic ValueError to a more specific DataItemValidationError
            # to provide clearer context when setting default values fails
            raise DataItemValidationError(
                instance,
                self,
                str(exc.__cause__ or exc),
                operation=f"Setting default value ({instance.__class__.__name__})",
            ) from exc

    def accept(self, visitor: object) -> None:
        """This is the visitor pattern's accept function.
        It calls the corresponding visit_MYCLASS method
        of the visitor object.

        Python's allow a generic base class implementation of
        this method so there's no need to write an accept function
        for each derived class unless you need to override the
        default behavior

        Args:
            visitor (object)
        """
        funcname = "visit_" + self.__class__.__name__
        func = getattr(visitor, funcname)
        func(self)

    def __set__(self, instance: Any, value: Any) -> None:
        """Set data item's value

        Args:
            instance (Any): instance of the DataSet
            value (Any): value to set
        """
        # Check if this item is computed (read-only)
        computed_prop = self.get_prop("data", "computed", None)
        if isinstance(computed_prop, ComputedProp):
            raise ValueError(f"Computed item '{self.get_name()}' is read-only")

        self._set_value_with_validation(instance, value, force_allow_none=False)

    def _set_value_with_validation(
        self, instance: Any, value: Any, force_allow_none: bool = False
    ) -> None:
        """Internal method to set data item's value with validation

        Args:
            instance (Any): instance of the DataSet
            value (Any): value to set
            force_allow_none (bool): if True, allow None values even when
             allow_none is False (used for default values)
        """
        vmode = get_validation_mode()

        # Early exit if validation is disabled
        if vmode == ValidationMode.DISABLED:
            setattr(instance, f"_{self._name}", value)
            return

        # Check if validation should be skipped for None values
        if value is None:
            if force_allow_none or self.get_prop("data", "allow_none", False):
                setattr(instance, f"_{self._name}", value)
                return

        # Perform validation
        try:
            self.check_value(value, raise_exception=True)
        except NotImplementedError:
            # Checking is not implemented for this item
            pass
        except Exception as exc:
            if vmode == ValidationMode.ENABLED:
                # Show a warning in replacement of the exception
                msg = f"Checking {instance.__class__.__name__}.{str(self)}: {exc}"
                warnings.warn(msg, DataItemValidationWarning)
            elif vmode == ValidationMode.STRICT:
                # Raise an exception if strict validation is enabled
                raise DataItemValidationError(instance, self, str(exc)) from exc
            else:
                raise ValueError(f"Unknown validation mode: {vmode}") from exc

        setattr(instance, f"_{self._name}", value)

    def __get__(self, instance: Any, klass: type) -> Any | None:
        if instance is not None:
            # Check if this item is computed
            computed_prop = self.get_prop("data", "computed", None)
            if isinstance(computed_prop, ComputedProp):
                # For computed items, calculate the value using the computed property
                return computed_prop(instance, self, None)
            else:
                # For regular items, return the stored value or default
                return getattr(instance, "_%s" % (self._name), self._default)
        else:
            return self

    def get_value(self, instance: Any) -> Any:
        """Return data item's value

        Args:
            instance (Any): instance of the DataSet

        Returns:
            Any: data item's value
        """
        return self.__get__(instance, instance.__class__)

    def check_item(self, instance: Any) -> bool:
        """Check data item's current value (calling method check_value)

        Args:
            instance (Any): instance of the DataSet

        Returns:
            Any: data item's value
        """
        value = getattr(instance, "_%s" % (self._name))
        return self.check_value(value)

    def check_value(self, value: Any, raise_exception: bool = False) -> bool:
        """Check if `value` is valid for this data item

        Args:
            value (Any): value to check
            raise_exception (bool): if True, raise an exception if the value is invalid.
            Defaults to True.

        Returns:
            bool: value
        """
        raise NotImplementedError()

    def from_string(self, string_value: str) -> Any:
        """Transform string into valid data item's value

        Args:
            string_value (str): string value

        Returns:
            Any: data item's value
        """
        raise NotImplementedError()

    def bind(self, instance: DataSet) -> DataItemVariable:
        """Return a DataItemVariable instance bound to the data item

        Args:
            instance (DataSet): instance of the DataSet

        Returns:
            DataItemVariable: DataItemVariable instance
        """
        return DataItemVariable(self, instance)

    def serialize(
        self,
        instance: DataSet,
        writer: HDF5Writer | JSONWriter | INIWriter,
    ) -> None:
        """Serialize this item using the writer object

        This is a default implementation that should work for
        everything but new datatypes

        Args:
            instance (DataSet): instance of the DataSet
            writer (HDF5Writer | JSONWriter | INIWriter): writer object
        """
        value = self.get_value(instance)
        writer.write(value)

    def get_value_from_reader(self, reader: HDF5Reader | JSONReader | INIReader) -> Any:
        """Reads value from the reader object, inside the try...except
        statement defined in the base item `deserialize` method

        This method is reimplemented in some child classes

        Args:
            reader (HDF5Reader | JSONReader | INIReader): reader object
        """
        return reader.read_any()

    def deserialize(
        self,
        instance: Any,
        reader: HDF5Reader | JSONReader | INIReader,
    ) -> None:
        """Deserialize this item using the reader object

        Default base implementation supposes the reader can
        detect expected datatype from the stream

        Args:
            instance (Any): instance of the DataSet
            reader (HDF5Reader | JSONReader | INIReader): reader object
        """
        try:
            value = self.get_value_from_reader(reader)
        except KeyError:
            self.set_default(instance)
            return
        except RuntimeError as e:
            if DEBUG_DESERIALIZE:
                import traceback

                print("DEBUG_DESERIALIZE enabled in datatypes.py", file=sys.stderr)
                traceback.print_stack()
                print(e, file=sys.stderr)
            self.set_default(instance)
            return
        self.__set__(instance, value)


class Obj:
    """An object that helps build default instances for
    ObjectItems"""

    def __init__(self, **kwargs) -> None:
        self.__dict__.update(kwargs)


class ObjectItem(DataItem):
    """Simple helper class implementing default
    for composite objects"""

    klass: type | None = None

    def set_default(self, instance: DataSet) -> None:
        """Make a copy of the default value

        Args:
            instance (DataSet): instance of the DataSet
        """
        # Avoid circular import
        # pylint: disable=import-outside-toplevel
        from guidata.dataset.conv import update_dataset

        if self.klass is not None:
            value = self.klass()  # pylint: disable=not-callable
            if self._default is not None:
                update_dataset(value, self._default)
            self.__set__(instance, value)

    def deserialize(
        self,
        instance: DataSet,
        reader: HDF5Reader | JSONReader | INIReader,
    ) -> None:
        """Deserialize this item using the reader object

        We build a new default object and deserialize it

        Args:
            instance (DataSet): instance of the DataSet
            reader (HDF5Reader | JSONReader | INIReader): reader object
        """
        if self.klass is not None:
            value = self.klass()  # pylint: disable=not-callable
            value.deserialize(reader)
            self.__set__(instance, value)


class DataItemProxy:
    """
    Proxy for DataItem objects

    This class is needed to construct GroupItem class
    (see module guidata.qtwidgets)

    Args:
        item (DataItem): data item to proxy
    """

    def __init__(self, item: DataItem):
        self.item = item

    def __str__(self):
        return self.item._name + "_proxy: " + self.__class__.__name__

    def get_name(self) -> str:
        """DataItem method proxy

        Returns:
            str: name
        """
        return self.item.get_name()

    def get_help(self, instance: DataSet) -> str:
        """DataItem method proxy

        Args:
            instance (DataSet): instance of the DataSet

        Returns:
            str: help string
        """
        return self.item.get_help(instance)

    def get_auto_help(self, instance: DataSet) -> str:
        """DataItem method proxy

        Args:
            instance (DataSet): instance of the DataSet

        Returns:
            str: help string
        """
        return self.item.get_auto_help(instance)

    def get_string_value(self, instance: DataSet) -> str:
        """DataItem method proxy

        Args:
            instance (DataSet): instance of the DataSet

        Returns:
            str: string value
        """
        return self.item.get_string_value(instance)

    def set_from_string(self, instance: DataSet, string_value: str) -> None:
        """DataItem method proxy

        Args:
            instance (DataSet): instance of the DataSet
            string_value (str): string value
        """
        self.item.set_from_string(instance, string_value)

    def set_default(self, instance: DataSet) -> None:
        """DataItem method proxy

        Args:
            instance (DataSet): instance of the DataSet
        """
        self.item.set_default(instance)

    def __set__(self, instance: Any, value: Any):
        pass

    def accept(self, visitor: object) -> None:
        """DataItem method proxy

        Args:
            visitor (object): visitor object
        """
        self.item.accept(visitor)

    def get_value(self, instance: DataItem) -> Any:
        """DataItem method proxy

        Args:
            instance (DataItem): instance of the DataItem

        Returns:
            Any: value
        """
        return self.item.get_value(instance)

    def check_item(self, instance: DataItem) -> bool:
        """DataItem method proxy

        Args:
            instance (DataItem): instance of the DataItem

        Returns:
            Any: value
        """
        return self.item.check_item(instance)

    def check_value(self, value: Any, raise_exception: bool = False) -> bool:
        """DataItem method proxy

        Args:
            value (Any): value
            raise_exception (bool): if True, raise an exception if the value is invalid.
             Defaults to True.

        Returns:
            Any: value
        """
        return self.item.check_value(value, raise_exception)

    def from_string(self, string_value: str) -> Any:
        """DataItem method proxy

        Args:
            string_value (str): string value

        Returns:
            Any: value
        """
        return self.item.from_string(string_value)

    def get_prop(self, realm: str, name: str, default=NoDefault) -> Any:
        """DataItem method proxy

        Args:
            realm (str): realm
            name (str): name
            default (Any): default value

        Returns:
            Any: value
        """
        return self.item.get_prop(realm, name, default)

    def get_prop_value(
        self, realm, instance: DataSet, name: str, default: Any = NoDefault
    ) -> Any:
        """DataItem method proxy

        Args:
            realm (str): realm
            instance (DataSet): instance of the DataSet
            name (str): name
            default (Any): default value

        Returns:
            Any: value
        """
        return self.item.get_prop_value(realm, instance, name, default)

    def set_prop(self, realm: str, **kwargs) -> DataItem:
        """DataItem method proxy

        Args:
            realm (str): realm
            kwargs: keyword arguments

        Returns:
            DataItem: data item
        """  # noqa
        return self.item.set_prop(realm, **kwargs)

    def bind(self, instance: DataSet) -> DataItemVariable:
        """DataItem method proxy

        Args:
            instance (DataSet): instance of the DataSet

        Returns:
            DataItemVariable: data item variable
        """
        return DataItemVariable(self, instance)


class DataItemVariable:
    """An instance of a DataItemVariable represent a binding between
    an item and a dataset.

    could be called a bound property.

    since DataItem instances are class attributes they need to have a
    DataSet instance to store their value. This class binds the two
    together.

    Args:
        item (DataItem): data item
        instance (DataSet): instance of the DataSet
    """

    def __init__(
        self,
        item: DataItem,
        instance: DataSet,
    ):
        self.item = item
        self.instance = instance

    def get_prop_value(self, realm: str, name: str, default: object = NoDefault) -> Any:
        """DataItem method proxy

        Args:
            realm (str): realm
            name (str): name
            default (object): default value

        Returns:
            Any: value
        """
        return self.item.get_prop_value(realm, self.instance, name, default)

    def get_prop(self, realm: str, name: str, default: type | None = NoDefault) -> Any:
        """DataItem method proxy

        Args:
            realm (str): realm
            name (str): name
            default (type | None): default value

        Returns:
            Any: value
        """
        return self.item.get_prop(realm, name, default)

    def get_help(self) -> str:
        """Re-implement DataItem method

        Returns:
            str: help string
        """
        return self.item.get_help(self.instance)

    def get_auto_help(self) -> str:
        """Re-implement DataItem method

        Returns:
            str: help string
        """
        return self.item.get_auto_help(self.instance)

    def get_string_value(self) -> str:
        """Return a unicode representation of the item's value
        obeying 'display' or 'repr' properties

        Returns:
            str: string value
        """
        return self.item.get_string_value(self.instance)

    def set_default(self) -> None:
        """Re-implement DataItem method"""
        return self.item.set_default(self.instance)

    def get(self) -> Any:
        """Re-implement DataItem method

        Returns:
            Any: value
        """
        return self.item.get_value(self.instance)

    def set(self, value: Any) -> None:
        """Re-implement DataItem method

        Args:
            value (Any): value
        """
        return self.item.__set__(self.instance, value)

    def set_from_string(self, string_value) -> None:
        """Re-implement DataItem method

        Args:
            string_value (str): string value
        """
        return self.item.set_from_string(self.instance, string_value)

    def check_item(self) -> bool:
        """Re-implement DataItem method

        Returns:
            Any: value
        """
        return self.item.check_item(self.instance)

    def check_value(self, value: Any, raise_exception: bool = False) -> bool:
        """Re-implement DataItem method

        Args:
            value (Any): value
            raise_exception (bool): if True, raise an exception if the value is invalid.
             Defaults to True.

        Returns:
            Any: value
        """
        return self.item.check_value(value, raise_exception)

    def from_string(self, string_value: str) -> Any:
        """Re-implement DataItem method

        Args:
            string_value (str): string value

        Returns:
            Any: value
        """
        return self.item.from_string(string_value)

    def label(self) -> str:
        """Re-implement DataItem method

        Returns:
            str: label
        """
        return self.item.get_prop("display", "label")


def collect_items_in_bases_order(
    bases: tuple[type], seen: set | None = None
) -> list[DataItem]:
    """Collect items in bases order

    Args:
        bases: tuple of base classes
        seen: set of already seen items to avoid duplicates

    Returns:
        List of data items
    """
    if seen is None:
        seen = set()

    # Collect all bases, respecting the order they appear in class definition
    # For DerivedAB(DatasetA, DatasetB), we want DatasetA items first
    all_bases = []
    for base in bases:
        # Get all bases of this base class in MRO order (excluding object)
        base_mro = [cls for cls in base.__mro__ if cls is not object]
        all_bases.extend(base_mro)

    # Remove duplicates while preserving order (use dict as ordered set)
    unique_bases = list(dict.fromkeys(all_bases))

    # Build a mapping of overridden items (most specific class wins)
    items_dict = {}  # name -> DataItem mapping
    for base in reversed(unique_bases):  # Reverse MRO: most specific first
        base_dict = getattr(base, "__dict__", {})
        for name, obj in base_dict.items():
            if isinstance(obj, DataItem):
                items_dict[name] = obj  # Later (more specific) definitions override

    # Now collect items following the inheritance chain order in bases
    # Process each base and its ancestors in the order they appear
    final_items = []
    processed_bases = set()

    for base in bases:  # Process in the order bases are listed in class definition
        # Get MRO for this base, but exclude DataSet and object
        base_chain = [
            cls
            for cls in base.__mro__
            if cls is not object and issubclass(cls, DataSet)
        ]

        # Process from most ancestral to most derived within this chain
        for cls in reversed(base_chain):
            if cls in processed_bases:
                continue  # Skip if already processed
            processed_bases.add(cls)

            cls_dict = getattr(cls, "__dict__", {})
            for name, obj in cls_dict.items():
                if isinstance(obj, DataItem) and name not in seen:
                    final_items.append(items_dict[name])  # Use the overridden version
                    seen.add(name)

    return final_items


class DataSetMeta(type):
    """
    DataSet metaclass

    Create class attribute `_items`: list of the DataSet class attributes,
    created in the same order as these attributes were written
    """

    def __new__(
        cls: type, name: str, bases: Any, dct: dict[str, Any], **kwargs
    ) -> type:
        # Filter out DataSet configuration kwargs (handled by __init_subclass__)
        # These include: title, comment, icon, readonly
        items = {item._name: item for item in collect_items_in_bases_order(bases)}
        for attrname, value in list(dct.items()):
            if isinstance(value, DataItem):
                value.set_name(attrname)
                if attrname in items:
                    value._order = items[attrname]._order
                items[attrname] = value
        dct["_items"] = list(items.values())
        # Pass kwargs through to type.__new__ (which will trigger __init_subclass__)
        return super().__new__(cls, name, bases, dct, **kwargs)


Meta_Py3Compat = DataSetMeta("Meta_Py3Compat", (object,), {})


AnyDataSet = TypeVar("AnyDataSet", bound="DataSet")


def assert_datasets_equal(ds1: DataSet, ds2: DataSet, msg: str | None = None) -> None:
    """Assert that two DataSet objects are equal. Exception message shows details on
    which attributes are different, if any.

    Args:
        ds1: The first DataSet to compare.
        ds2: The second DataSet to compare.
        msg: Optional message to include in the assertion error.

    Raises:
        AssertionError: If the DataSet objects are not equal.
    """
    diff: list[str] = []
    items1, items2 = ds1.get_items(), ds2.get_items()
    if len(items1) != len(items2):
        diff.append(f"Different number of items: {len(items1)} != {len(items2)}")
    else:
        for item1, item2 in zip(items1, items2):
            name1, name2 = item1.get_name(), item2.get_name()
            if item1 != item2:
                diff.append(f"Item '{name1}' differs from '{name2}'")
            val1, val2 = item1.get_value(ds1), item2.get_value(ds2)
            diffval = False
            if isinstance(val1, (list, tuple, set)):
                if not all(v1 == v2 for v1, v2 in zip(val1, val2)):
                    diffval = True
            elif isinstance(val1, np.ndarray) and isinstance(val2, np.ndarray):
                if not np.array_equal(val1, val2):
                    diffval = True
            elif val1 != val2:
                diffval = True
            if diffval:
                diff.append(f"Item '{name1}' has different value: {val1} != {val2}")
    if diff:
        msg = "" if msg is None else msg + "\n"
        raise AssertionError(msg + "Datasets are not equal:\n" + "\n".join(diff))


class DataSet(metaclass=DataSetMeta):
    """Construct a DataSet object is a set of DataItem objects

    Args:
        title (str): title
        comment (str): comment. Text shown on the top of the first data item
        icon (str): icon filename as in image search paths
        readonly (bool): if True, the DataSet is read-only
        skip_defaults (bool): if True, do not set default values for items
    """

    _items: list[DataItem] = []
    __metaclass__ = DataSetMeta  # keep it even with Python 3 (see DataSetMeta)

    # Class-level configuration (set via __init_subclass__)
    _class_title: str | None = None
    _class_comment: str | None = None
    _class_icon: str = ""
    _class_readonly: bool = False

    def __init_subclass__(
        cls,
        title: str | None = None,
        comment: str | None = None,
        icon: str = "",
        readonly: bool = False,
        **kwargs,
    ) -> None:
        """Called when a class inherits from DataSet.

        This allows configuring DataSet metadata in the class definition:

            class MyParams(DataSet, title="My Parameters", icon="myicon.png"):
                param1 = FloatItem("Parameter 1")

        Args:
            title: Default title for this DataSet class
            comment: Default comment for this DataSet class
            icon: Default icon for this DataSet class
            readonly: Default readonly state for this DataSet class
            **kwargs: Additional arguments passed to parent __init_subclass__
        """
        super().__init_subclass__(**kwargs)
        cls._class_title = title
        cls._class_comment = comment
        cls._class_icon = icon
        cls._class_readonly = readonly

    def __init__(
        self,
        title: str | None = None,
        comment: str | None = None,
        icon: str = "",
        readonly: bool = False,
        skip_defaults: bool = False,
    ):
        # Priority: instance parameter > class-level config > empty/computed default
        self.__icon = icon if icon else self._class_icon
        self.__readonly: bool = readonly or self._class_readonly

        # Handle title: instance param > class config > docstring fallback
        if title is not None:
            # Explicitly passed to __init__ (even if empty string)
            self.__title = title
        elif self._class_title is not None:
            # Set at class level via __init_subclass__ (even if empty string)
            self.__title = self._class_title
        else:
            # Fall back to docstring (for backward compatibility)
            comp_title, comp_comment = self._compute_title_and_comment()
            self.__title = comp_title

        # Handle comment: instance param > class config > None (no docstring fallback)
        if comment is not None:
            # Explicitly passed to __init__
            self.__comment = comment
        elif self._class_comment is not None:
            # Set at class level via __init_subclass__
            self.__comment = self._class_comment
        else:
            # No automatic fallback to docstring for comment
            self.__comment = None

        if not skip_defaults:
            self.set_defaults()

    def get_items(self, copy=False) -> list[DataItem]:
        """Returns all the DataItem objects from the DataSet instance. Ignore private
        items that have a name starting with an underscore (e.g. '_private_item = ...')

        Args:
            copy: If True, deepcopy the DataItem list, else return the original.
            Defaults to False.

        Returns:
            _description_
        """
        result_items = self._items if not copy else deepcopy(self._items)
        return list(filter(lambda s: not s.get_name().startswith("_"), result_items))

    @classmethod
    def create(cls: type[AnyDataSet], **kwargs) -> AnyDataSet:
        """Create a new instance of the DataSet class

        Args:
            kwargs: keyword arguments to set the DataItem values

        Returns:
            DataSet instance
        """  # noqa
        # Validate that all kwargs correspond to actual items in the class
        for name in kwargs:
            for item in cls._items:
                if item._name == name:
                    break
            else:
                raise AttributeError(
                    f"DataSet class '{cls.__name__}' has no attribute '{name}'"
                )

        # Create the instance but skip set_defaults in __init__
        instance = cls(skip_defaults=True)

        # Set only the defaults for items that don't have values in kwargs
        for item in instance._items:
            if item._name not in kwargs:
                item.set_default(instance)

        # Now set the provided values
        for name, value in kwargs.items():
            setattr(instance, name, value)

        return instance

    def _get_translation(self):
        """We try to find the translation function (_) from the module
        this class was created in

        This function is unused but could be useful to translate strings that
        cannot be translated at the time they are created.
        """
        module = sys.modules[self.__class__.__module__]
        if hasattr(module, "_"):
            return module._
        else:
            return lambda x: x

    def _compute_title_and_comment(self) -> tuple[str, str | None]:
        """
        Private method to compute title and comment of the data set from docstring.

        Returns tuple of (title, comment) where:
        - title: first line of docstring (stripped), or class name if no docstring
        - comment: remaining lines of docstring (stripped), or None if no lines
        """
        comp_title = self.__class__.__name__
        comp_comment = None
        if self.__doc__:
            doc_lines = self.__doc__.splitlines()
            # Remove empty lines at the beginning
            while doc_lines and not doc_lines[0].strip():
                del doc_lines[0]
            if doc_lines:
                # First line becomes the title
                comp_title = doc_lines.pop(0).strip()
            if doc_lines:
                # Remaining lines become the comment
                comp_comment = "\n".join([x.strip() for x in doc_lines])
        return comp_title, comp_comment

    def get_title(self) -> str:
        """Return data set title

        Returns:
            str: title
        """
        return self.__title

    def set_title(self, title: str) -> None:
        """Set data set title

        Args:
            title (str): title
        """
        self.__title = title

    def get_comment(self) -> str | None:
        """Return data set comment

        Returns:
            str | None: comment
        """
        return self.__comment

    def set_comment(self, comment: str | None) -> None:
        """Set data set comment

        Args:
            comment (str | None): comment
        """
        self.__comment = comment

    def get_icon(self) -> str | None:
        """Return data set icon

        Returns:
            str | None: icon
        """
        return self.__icon

    def set_icon(self, icon: str) -> None:
        """Set data set icon

        Args:
            icon (str): icon
        """
        self.__icon = icon

    def set_defaults(self) -> None:
        """Set default values"""
        for item in self._items:
            item.set_default(self)

    def __str__(self) -> str:
        """Return string representation of the data set"""
        return self.to_string(debug=False)

    def check(self) -> list[str]:
        """Check the dataset item values

        Returns:
            list[str]: list of errors
        """
        errors = []
        for item in self._items:
            if not item.check_item(self) and item._name is not None:
                errors.append(item._name)
        return errors

    def text_edit(self) -> None:
        """Edit data set with text input only"""
        from guidata.dataset import textedit

        self.accept(textedit.TextEditVisitor(self))

    def edit(
        self,
        parent: QWidget | None = None,
        apply: Callable | None = None,
        wordwrap: bool = True,
        size: QSize | tuple[int, int] | None = None,
        object_name: str | None = None,
    ) -> int:
        """Open a dialog box to edit data set

        Args:
            parent: parent widget (default is None, meaning no parent)
            apply: apply callback (default is None)
            wordwrap: if True, comment text is wordwrapped
            size: dialog size (QSize object or integer tuple (width, height))
            object_name: object name for the dialog (default is None, meaning
             use class name + "Dialog")

        Returns:
            Dialog exit code.
        """
        # Importing those modules here avoids Qt dependency when
        # guidata is used without Qt
        # pylint: disable=import-outside-toplevel
        from guidata.dataset.qtwidgets import DataSetEditDialog
        from guidata.qthelpers import exec_dialog

        dlg = DataSetEditDialog(
            self,
            icon=self.__icon,
            parent=parent,
            apply=apply,
            wordwrap=wordwrap,
            size=size,
        )
        dlg.setObjectName(object_name or self.__class__.__name__ + "Dialog")

        return exec_dialog(dlg)

    def view(
        self,
        parent: QWidget | None = None,
        wordwrap: bool = True,
        size: QSize | tuple[int, int] | None = None,
    ) -> None:
        """Open a dialog box to view data set

        Args:
            parent: parent widget (default is None, meaning no parent)
            wordwrap: if True, comment text is wordwrapped
            size: dialog size (QSize object or integer tuple (width, height))
        """
        # Importing those modules here avoids Qt dependency when
        # guidata is used without Qt
        # pylint: disable=import-outside-toplevel
        from guidata.dataset.qtwidgets import DataSetShowDialog
        from guidata.qthelpers import exec_dialog

        dial = DataSetShowDialog(
            self, icon=self.__icon, parent=parent, wordwrap=wordwrap, size=size
        )
        return exec_dialog(dial)

    def is_readonly(self) -> bool:
        return bool(self.__readonly)

    def set_readonly(self, readonly: bool = True):
        self.__readonly = readonly

    def _get_items_for_text_representation(self) -> list[DataItem]:
        """Get items for text representation, excluding trailing separators

        Returns:
            list: List of items with trailing separators filtered out
        """
        filtered_items = list(self._items)
        while filtered_items and isinstance(filtered_items[-1], SeparatorItem):
            filtered_items.pop()
        return filtered_items

    def to_string(
        self,
        debug: bool | None = False,
        indent: str | None = None,
        align: bool | None = False,
        show_hidden: bool | None = True,
    ) -> str:
        """Return readable string representation of the data set
        If debug is True, add more details on data items

        Args:
            debug (bool): if True, add more details on data items
            indent (str): indentation string (default is None,
                meaning no indentation)
            align (bool): if True, align data items (default is False)
            show_hidden (bool): if True, show hidden data items
                (default is True)

        Returns:
            str: string representation of the data set
        """
        if indent is None:
            indent = "\n    "
        txt = "%s:" % (self.__title)

        def _get_label(item):
            if debug:
                return item._name
            else:
                return item.get_prop_value("display", self, "label")

        # Get items for text representation (excluding trailing separators)
        filtered_items = self._get_items_for_text_representation()

        length = 0
        if align:
            for item in filtered_items:
                item_length = len(_get_label(item))
                if item_length > length:
                    length = item_length
        for item in filtered_items:
            try:
                hide = item.get_prop_value("display", self, "hide")
                if not show_hidden and hide is True:
                    continue
            except KeyError:
                pass
            if isinstance(item, ObjectItem):
                composite_dataset = item.get_value(self)
                txt += indent + composite_dataset.to_string(
                    debug=debug, indent=indent + "  "
                )
                continue
            elif isinstance(item, BeginGroup):
                txt += "%s%s:" % (indent, item._name)
                indent += "  "
                continue
            elif isinstance(item, EndGroup):
                indent = indent[:-2]
                continue
            value = getattr(self, "_%s" % (item._name))
            value_str = "-" if value is None else item.get_string_value(self)
            if debug:
                label = item._name
            else:
                label = item.get_prop_value("display", self, "label")
                if not label:
                    # For BoolItem without label, use text as label
                    label = item.get_prop_value("display", self, "text", "")
            if length and label is not None:
                label = label.ljust(length)
            txt += f"{indent}{label}: {value_str}"
            if debug:
                txt += " (" + item.__class__.__name__ + ")"
        return txt

    def accept(self, vis: object) -> None:
        """Helper function that passes the visitor to the accept methods of all
        the items in this dataset

        Args:
            vis (object): visitor object
        """
        for item in self._items:
            item.accept(vis)

    def to_html(self) -> str:
        """Return HTML representation of the dataset.

        Similar to Sigima's TableResult transpose format with:

        - Title and comment in blue
        - Two-column table: item names (right-aligned) and values (left-aligned)
        - For BoolItem: checkbox in first column with label:text formatting

        Returns:
            HTML representation
        """
        # Create the title with comment
        # Use a lighter blue (#5294e2) that works well in both light and dark modes
        html = f'<u><b style="color: #5294e2">{self.__title}</b></u>:'

        if self.__comment:
            # Add comment on new line, also in blue
            html += f'<br><span style="color: #5294e2">{self.__comment}</span>'

        # Get items for representation (excluding trailing separators)
        filtered_items = self._get_items_for_text_representation()

        if not filtered_items:
            html += "<br><em>No items to display</em>"
            return html

        # Start table
        html += '<table border="0">'

        for item in filtered_items:
            # Skip group items and separators for HTML representation
            if isinstance(item, (BeginGroup, EndGroup, SeparatorItem)):
                continue

            # Skip hidden items
            try:
                hide = item.get_prop_value("display", self, "hide")
                if hide is True:
                    continue
            except KeyError:
                pass

            # Handle ObjectItem (nested datasets)
            if isinstance(item, ObjectItem):
                composite_dataset = item.get_value(self)
                if hasattr(composite_dataset, "to_html"):
                    item_label = item.get_prop_value("display", self, "label")
                    html += (
                        f'<tr><td style="text-align: right; vertical-align: top;">'
                        f"{item_label}:</td>"
                    )
                    nested_html = composite_dataset.to_html(transpose=False)
                    html += (
                        f'<td style="text-align: left; padding-left: 10px;">'
                        f"{nested_html}</td></tr>"
                    )
                continue

            # Get item label and value
            label = item.get_prop_value("display", self, "label")
            if not label:
                # For BoolItem without label, use name:text formatting
                label = item.get_prop_value("display", self, "text", "")

            # Get string representation of value
            value_str = item.get_string_value(self)

            html += (
                f'<tr><td style="text-align: right; vertical-align: top;">{label}:</td>'
            )
            html += (
                f'<td style="text-align: left; padding-left: 10px;">'
                f"{value_str}</td></tr>"
            )

        html += "</table>"
        return html

    def serialize(self, writer: HDF5Writer | JSONWriter | INIWriter) -> None:
        """Serialize the dataset

        Args:
            writer (HDF5Writer | JSONWriter | INIWriter): writer object
        """
        for item in self._items:
            with writer.group(item._name):
                item.serialize(self, writer)

    def deserialize(self, reader: HDF5Reader | JSONReader | INIReader) -> None:
        """Deserialize the dataset

        Args:
            reader (HDF5Reader | JSONReader | INIReader): reader object
        """
        for item in self._items:
            with reader.group(item._name):
                if item.get_prop("data", "computed", None) is not None:
                    continue  # Skip computed items
                try:
                    item.deserialize(self, reader)
                except RuntimeError as error:
                    if DEBUG_DESERIALIZE:
                        import traceback

                        print(
                            "DEBUG_DESERIALIZE enabled in datatypes.py", file=sys.stderr
                        )
                        traceback.print_stack()
                        print(error, file=sys.stderr)
                    item.set_default(self)

    def read_config(self, conf: UserConfig, section: str, option: str) -> None:
        """Read configuration from a UserConfig instance

        Args:
            conf (UserConfig): UserConfig instance
            section (str): section name
            option (str): option name
        """
        reader = INIReader(conf, section, option)
        self.deserialize(reader)

    def write_config(self, conf: UserConfig, section: str, option: str) -> None:
        """Write configuration to a UserConfig instance

        Args:
            conf (UserConfig): UserConfig instance
            section (str): section name
            option (str): option name
        """
        writer = INIWriter(conf, section, option)
        self.serialize(writer)

    @classmethod
    def set_global_prop(klass, realm: str, **kwargs) -> None:
        """Set global properties for all data items in the dataset

        Args:
            realm (str): realm name
            kwargs (dict): properties to set
        """  # noqa
        for item in klass._items:
            item.set_prop(realm, **kwargs)


class ActivableDataSet(DataSet):
    """An ActivableDataSet instance must have an "enable" class attribute which
    will set the active state of the dataset instance
    (see example in: tests/activable_dataset.py)

    Args:
        title (str): dataset title (optional)
        comment (str): dataset comment (optional)
        icon (str): dataset icon. Default is "" (no icon)
    """

    _activable = True  # default *instance* attribute value
    _active = True
    _activable_prop = GetAttrProp("_activable")
    _active_prop = GetAttrProp("_active")

    @property
    @abstractmethod
    def enable(self) -> DataItem: ...

    def __init__(
        self,
        title: str | None = None,
        comment: str | None = None,
        icon: str = "",
    ):
        DataSet.__init__(self, title, comment, icon)

    @classmethod
    def active_setup(cls) -> None:
        """
        This class method must be called after the child class definition
        in order to setup the dataset active state
        """
        cls.set_global_prop("display", active=cls._active_prop)
        cls.enable.set_prop(  # type:ignore
            "display", active=True, hide=cls._activable_prop, store=cls._active_prop
        )

    def set_activable(self, activable: bool):
        self._activable = not activable
        self._active = self.enable


class DataSetGroup:
    """Construct a DataSetGroup object, used to group several datasets together

    Args:
        datasets (list[DataSet]): list of datasets
        title (str): group title (optional)
        icon (str): group icon. Default is "" (no icon)

    This class tries to mimics the DataSet interface.

    The GUI should represent it as a notebook with one page for each
    contained dataset.
    """

    ALLOWED_MODES = ("tabs", "table", None)

    def __init__(
        self,
        datasets: list[DataSet],
        title: str | None = None,
        icon: str = "",
    ) -> None:
        self.__icon = icon
        self.datasets = datasets
        if title:
            self.__title = title
        else:
            self.__title = self.__class__.__name__

    def __str__(self) -> str:
        return "\n".join([dataset.__str__() for dataset in self.datasets])

    def get_title(self) -> str:
        """Return data set group title

        Returns:
            str: data set group title
        """
        return self.__title

    def get_comment(self) -> None:
        """Return data set group comment --> not implemented (will return None)

        Returns:
            None: data set group comment
        """
        return None

    def get_icon(self) -> str | None:
        """Return data set icon

        Returns:
            str | None: data set icon
        """
        return self.__icon

    def check(self) -> list[list[str]]:
        """Check data set group items

        Returns:
            list[list[str]]: list of errors
        """
        return [dataset.check() for dataset in self.datasets]

    def text_edit(self) -> None:
        """Edit data set with text input only"""
        raise NotImplementedError()

    def edit(
        self,
        parent: QWidget | None = None,
        apply: Callable | None = None,
        wordwrap: bool = True,
        size: QSize | tuple[int, int] | None = None,
        mode: str | None = None,
    ) -> int:
        """Open a dialog box to edit data set

        Args:
            parent: parent widget. Defaults to None.
            apply: apply callback. Defaults to None.
            wordwrap: if True, comment text is wordwrapped
            size: dialog size (default: None)
            mode: (str): dialog window style to use. Allowed values are "tabs",
             "table" and None. Use "tabs" to navigate between datasets with tabs.
             Use "table" to create a table with one dataset by row (allows
             dataset editing by double clicking on a row). Defaults to None.

        Returns:
            int: dialog box return code
        """
        # Importing those modules here avoids Qt dependency when
        # guidata is used without Qt
        # pylint: disable=import-outside-toplevel

        assert mode in self.ALLOWED_MODES

        from guidata.dataset.qtwidgets import (
            DataSetGroupEditDialog,
            DataSetGroupTableEditDialog,
        )
        from guidata.qthelpers import exec_dialog

        dial: QDialog
        if mode in ("tabs", None):
            dial = DataSetGroupEditDialog(
                instance=self,  # type: ignore
                icon=self.__icon,
                parent=parent,
                apply=apply,
                wordwrap=wordwrap,
                size=size,
            )
            return exec_dialog(dial)
        else:
            dial = DataSetGroupTableEditDialog(
                instance=self,
                icon=self.__icon,
                parent=parent,
                apply=apply,
                wordwrap=wordwrap,
                size=size,
            )
            return exec_dialog(dial)

    def accept(self, vis: object) -> None:
        """Helper function that passes the visitor to the accept methods of all
        the items in this dataset

        Args:
            vis (object): visitor
        """
        for dataset in self.datasets:
            dataset.accept(vis)

    def is_readonly(self) -> bool:
        """Return True if all datasets in the DataSetGroup are in readonly mode.

        Returns:
            True if all datasets are in readonly, else False
        """
        return all((ds.is_readonly() for ds in self.datasets))

    def set_readonly(self, readonly=True):
        """Set all datasets of the dataset group to readonly mode

        Args:
            readonly: Readonly flag. Defaults to True.
        """
        _ = [d.set_readonly(readonly) for d in self.datasets]


class GroupItem(DataItemProxy):
    """GroupItem proxy

    Args:
        item (DataItem): data item
    """

    def __init__(self, item: DataItem) -> None:
        DataItemProxy.__init__(self, item)
        self.group: list[Any] = []


class BeginGroup(DataItem):
    """Data item which does not represent anything
    but a begin flag to define a data set group

    Args:
        label (str): group label
    """

    def __init__(self, label: str) -> None:
        super().__init__(label)

    def serialize(self, instance, writer) -> None:
        pass

    def deserialize(self, instance, reader) -> None:
        pass

    def get_group(self) -> "GroupItem":
        return GroupItem(self)


class EndGroup(DataItem):
    """Data item which does not represent anything
    but an end flag to define a data set group

    Args:
        label (str): group label
    """

    def __init__(self, label: str) -> None:
        super().__init__(label)

    def serialize(self, instance, writer) -> None:
        pass

    def deserialize(self, instance, reader) -> None:
        pass


class TabGroupItem(GroupItem):
    pass


class BeginTabGroup(BeginGroup):
    """Data item which does not represent anything
    but a begin flag to define a data set tab group

    Args:
        label (str): group label
    """

    def get_group(self) -> "TabGroupItem":
        return TabGroupItem(self)


class EndTabGroup(EndGroup):
    """Data item which does not represent anything
    but an end flag to define a data set tab group

    Args:
        label (str): group label
    """

    pass


class SeparatorItem(DataItem):
    """Data item which represents a visual separator between other items

    In textual representation, it appears as a series of dashes.
    In GUI, it appears as a horizontal gray line.

    Args:
        label (str): optional label for the separator (default: "")
    """

    def __init__(self, label: str = "") -> None:
        super().__init__(label)

    def get_string_value(self, instance: DataSet) -> str:
        """Return a formatted string representation of the separator

        Args:
            instance (DataSet): instance of the DataSet

        Returns:
            str: string representation as a series of dashes
        """
        return "-" * 50  # Return a line of 50 dashes

    def serialize(
        self, instance: DataSet, writer: HDF5Writer | JSONWriter | INIWriter
    ) -> None:
        """Serialize this item using the writer object

        Separators don't store any data, so this is a no-op.

        Args:
            instance (DataSet): instance of the DataSet
            writer (HDF5Writer | JSONWriter | INIWriter): writer object
        """
        pass

    def deserialize(
        self, instance: DataSet, reader: HDF5Reader | JSONReader | INIReader
    ) -> None:
        """Deserialize this item using the reader object

        Separators don't store any data, so this is a no-op.

        Args:
            instance (DataSet): instance of the DataSet
            reader (HDF5Reader | JSONReader | INIReader): reader object
        """
        pass

    def check_value(self, value: Any, raise_exception: bool = False) -> bool:
        """Check if `value` is valid for this data item

        Separators don't store values, so always return True.

        Args:
            value (Any): value to check
            raise_exception (bool): if True, raise an exception if the value is invalid

        Returns:
            bool: True (separators always have valid "values")
        """
        return True

    def from_string(self, string_value: str) -> None:
        """Transform string into valid data item's value

        Separators don't store values, so return None.

        Args:
            string_value (str): string value

        Returns:
            None: separators don't have values
        """
        return None

    def get_value(self, instance: DataSet) -> None:
        """Return data item's value

        Separators don't store values.

        Args:
            instance (DataSet): instance of the DataSet

        Returns:
            None: separators don't have values
        """
        return None

    def __set__(self, instance: DataSet, value: Any) -> None:
        """Set data item's value

        Separators don't store values, so this is a no-op.

        Args:
            instance (DataSet): instance of the DataSet
            value (Any): value to set (ignored)
        """
        pass  # Separators don't store any value



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)

# pylint: disable=C0103

"""
Miscellaneous utility functions
-------------------------------

Running programs
^^^^^^^^^^^^^^^^

.. autofunction:: run_program

.. autofunction:: is_program_installed

.. autofunction:: run_shell_command

Strings
^^^^^^^

.. autofunction:: to_string

.. autofunction:: decode_fs_string
"""

from __future__ import annotations

import collections.abc
import ctypes
import locale
import os
import os.path as osp
import subprocess
import sys
from typing import Any, Type

from guidata.userconfig import get_home_dir

# MARK: Strings, Locale ----------------------------------------------------------------


def to_string(obj: Any) -> str:
    """Convert object to string.

    If `obj` is bytes-like, try decoding as UTF-8, falling back to Latin-1.
    Otherwise, use str().
    """
    if isinstance(obj, (bytes, bytearray, memoryview)):
        try:
            return obj.decode("utf-8")
        except UnicodeDecodeError:
            return obj.decode("latin-1")
    return str(obj)


def decode_fs_string(string: str) -> str:
    """Convert string from file system charset to unicode

    Args:
        string (str): String to convert

    Returns:
        str: Converted string
    """
    charset = sys.getfilesystemencoding()
    if charset is None:
        charset = locale.getpreferredencoding()
    return string.decode(charset)


def get_system_lang() -> str | None:
    """
    Retrieves the system language name.

    This function uses `locale.getlocale()` to obtain the locale name based on
    the current user's settings. If that fails on Windows (e.g. for frozen
    applications), it uses the Win32 API function `GetUserDefaultLocaleName`
    to obtain the locale name.

    Returns:
        The locale name in a format like 'en_US', or None if the function fails
        to retrieve the locale name.
    """
    lang = locale.getlocale()[0]
    if lang is None and os.name == "nt":
        kernel32 = ctypes.WinDLL("kernel32", use_last_error=True)
        GetUserDefaultLocaleName = kernel32.GetUserDefaultLocaleName
        GetUserDefaultLocaleName.argtypes = [ctypes.c_wchar_p, ctypes.c_int]
        GetUserDefaultLocaleName.restype = ctypes.c_int
        locale_name = ctypes.create_unicode_buffer(85)
        if GetUserDefaultLocaleName(locale_name, 85):
            lang = locale_name.value.replace("-", "_")
    return lang


# MARK: Interface checking -------------------------------------------------------------


def assert_interface_supported(klass: Type, iface: Type) -> None:
    """Makes sure a class supports an interface

    Args:
        klass (Type): The class.
        iface (Type): The interface.

    Raises:
        AssertionError: If the class does not support the interface.
    """
    for name, func in list(iface.__dict__.items()):
        if name == "__inherits__":
            continue
        if isinstance(func, collections.abc.Callable):
            assert hasattr(klass, name), "Attribute %s missing from %r" % (name, klass)
            imp_func = getattr(klass, name)
            imp_code = imp_func.__code__
            code = func.__code__
            imp_nargs = imp_code.co_argcount
            nargs = code.co_argcount
            if imp_code.co_varnames[:imp_nargs] != code.co_varnames[:nargs]:
                assert False, "Arguments of %s.%s differ from interface: %r!=%r" % (
                    klass.__name__,
                    imp_func.__name__,
                    imp_code.co_varnames[:imp_nargs],
                    code.co_varnames[:nargs],
                )
        else:
            pass  # should check class attributes for consistency


def assert_interfaces_valid(klass: Type) -> None:
    """Makes sure a class supports the interfaces it declares

    Args:
        klass (Type): The class.

    Raises:
        AssertionError: If the class does not support the interfaces it declares.
    """
    assert hasattr(klass, "__implements__"), "Class doesn't implements anything"
    for iface in klass.__implements__:
        assert_interface_supported(klass, iface)
        if hasattr(iface, "__inherits__"):
            base = iface.__inherits__()
            assert issubclass(klass, base), "%s should be a subclass of %s" % (
                klass,
                base,
            )


# MARK: Module, scripts, programs ------------------------------------------------------


def get_module_path(modname: str) -> str:
    """Return module *modname* base path.

    Args:
        modname (str): The module name.

    Returns:
        str: The module base path.
    """
    module = sys.modules.get(modname, __import__(modname))
    return osp.abspath(osp.dirname(module.__file__))


def is_program_installed(basename: str) -> str | None:
    """Return program absolute path if installed in PATH, otherwise None.

    Args:
        basename (str): The program base name.

    Returns:
        str | None: The program absolute path if installed in PATH,
            otherwise None.
    """
    for path in os.environ["PATH"].split(os.pathsep):
        abspath = osp.join(path, basename)
        if osp.isfile(abspath):
            return abspath


def run_program(
    name, args: str = "", cwd: str = None, shell: bool = True, wait: bool = False
) -> None:
    """Run program in a separate process.

    Args:
        name (str): The program name.
        args (str): The program arguments. Defaults to "".""
        cwd (str): The current working directory. Defaults to None.
        shell (bool): If True, run program in a shell. Defaults to True.
        wait (bool): If True, wait for program to finish. Defaults to False.

    Raises:
        RuntimeError: If program is not installed.
    """
    path = is_program_installed(name)
    if not path:
        raise RuntimeError("Program %s was not found" % name)
    command = [path]
    if args:
        command.append(args)
    if wait:
        subprocess.call(" ".join(command), cwd=cwd, shell=shell)
    else:
        subprocess.Popen(" ".join(command), cwd=cwd, shell=shell)


class ProgramError(Exception):
    """Exception raised when a shell command failed to execute."""

    pass


def alter_subprocess_kwargs_by_platform(**kwargs):
    """
    Given a dict, populate kwargs to create a generally
    useful default setup for running subprocess processes
    on different platforms. For example, `close_fds` is
    set on posix and creation of a new console window is
    disabled on Windows.

    This function will alter the given kwargs and return
    the modified dict.
    """
    kwargs.setdefault("close_fds", os.name == "posix")
    if os.name == "nt":
        CONSOLE_CREATION_FLAGS = 0  # Default value
        # See: https://msdn.microsoft.com/en-us/library/windows/desktop/ms684863%28v=vs.85%29.aspx
        CREATE_NO_WINDOW = 0x08000000
        # We "or" them together
        CONSOLE_CREATION_FLAGS |= CREATE_NO_WINDOW
        kwargs.setdefault("creationflags", CONSOLE_CREATION_FLAGS)
    return kwargs


def run_shell_command(cmdstr, **subprocess_kwargs):
    """
    Execute the given shell command.

    Note that args and kwargs will be passed to the subprocess call.

    If 'shell' is given in subprocess_kwargs it must be True,
    otherwise ProgramError will be raised.

    If 'executable' is not given in subprocess_kwargs, it will
    be set to the value of the SHELL environment variable.

    Note that stdin, stdout and stderr will be set by default
    to PIPE unless specified in subprocess_kwargs.

    :str cmdstr: The string run as a shell command.
    :subprocess_kwargs: These will be passed to subprocess.Popen.
    """
    if "shell" in subprocess_kwargs and not subprocess_kwargs["shell"]:
        raise ProgramError(
            'The "shell" kwarg may be omitted, but if provided it must be True.'
        )
    else:
        subprocess_kwargs["shell"] = True

    if "executable" not in subprocess_kwargs:
        subprocess_kwargs["executable"] = os.getenv("SHELL")

    for stream in ["stdin", "stdout", "stderr"]:
        subprocess_kwargs.setdefault(stream, subprocess.PIPE)
    subprocess_kwargs = alter_subprocess_kwargs_by_platform(**subprocess_kwargs)
    return subprocess.Popen(cmdstr, **subprocess_kwargs)


# MARK: Path utils ---------------------------------------------------------------------


def getcwd_or_home():
    """Safe version of getcwd that will fallback to home user dir.

    This will catch the error raised when the current working directory
    was removed for an external program.
    """
    try:
        return os.getcwd()
    except OSError:
        print(
            "WARNING: Current working directory was deleted, "
            "falling back to home directory"
        )
        return get_home_dir()


def remove_backslashes(path):
    """Remove backslashes in *path*

    For Windows platforms only.
    Returns the path unchanged on other platforms.

    This is especially useful when formatting path strings on
    Windows platforms for which folder paths may contain backslashes
    and provoke unicode decoding errors in Python 3 (or in Python 2
    when future 'unicode_literals' symbol has been imported)."""
    if os.name == "nt":
        # Removing trailing single backslash
        if path.endswith("\\") and not path.endswith("\\\\"):
            path = path[:-1]
        # Replacing backslashes by slashes
        path = path.replace("\\", "/")
        path = path.replace("/'", "\\'")
    return path


# MARK: Date utils ---------------------------------------------------------------------


def convert_date_format(format_string: str) -> str:
    """
    Converts a date format string in Python strftime format to QDateTime style format.

    Args:
        format_string: The date format string in Python strftime format.

    Returns:
        The converted date format string in QDateTime style.

    Examples:
        >>> format_string = '%d.%m.%Y'
        >>> qt_format = convert_date_format(format_string)
        >>> print(qt_format)
        dd.MM.yyyy
    """
    format_mapping = {
        "%d": "dd",
        "%-d": "d",
        "%dd": "dd",
        "%-dd": "d",
        "%ddd": "ddd",
        "%dddd": "dddd",
        "%b": "MMM",
        "%B": "MMMM",
        "%m": "MM",
        "%-m": "M",
        "%mm": "MM",
        "%-mm": "M",
        "%y": "yy",
        "%Y": "yyyy",
        "%I": "h",
        "%H": "HH",
        "%-H": "H",
        "%M": "mm",
        "%-M": "m",
        "%S": "ss",
        "%-S": "s",
        "%z": "z",
        "%Z": "zzz",
    }

    qt_format = ""
    i = 0
    while i < len(format_string):
        if format_string[i : i + 2] in format_mapping:
            qt_format += format_mapping[format_string[i : i + 2]]
            i += 2
        elif format_string[i] in format_mapping:
            qt_format += format_mapping[format_string[i]]
            i += 1
        else:
            qt_format += format_string[i]
            i += 1

    return qt_format



# content of conftest.py

import os

import qtpy

import guidata
from guidata.env import execenv
from guidata.utils.gitreport import format_git_info_for_pytest, get_git_info_for_modules

# Turn on unattended mode for executing tests without user interaction
execenv.unattended = True
execenv.verbose = "quiet"


def pytest_addoption(parser):
    """Add custom command line options to pytest."""
    parser.addoption(
        "--show-windows",
        action="store_true",
        default=False,
        help="Display Qt windows during tests (disables QT_QPA_PLATFORM=offscreen)",
    )


def pytest_configure(config):
    """Configure pytest based on command line options."""
    if config.option.durations is None:
        config.option.durations = 10  # Default to showing 10 slowest tests
    if not config.getoption("--show-windows"):
        os.environ.setdefault("QT_QPA_PLATFORM", "offscreen")


def pytest_report_header(config):
    """Add additional information to the pytest report header."""
    qtbindings_version = qtpy.PYSIDE_VERSION
    if qtbindings_version is None:
        qtbindings_version = qtpy.PYQT_VERSION
    infolist = [
        f"guidata {guidata.__version__}, "
        f"{qtpy.API_NAME} {qtbindings_version} [Qt version: {qtpy.QT_VERSION}]",
    ]

    # Git information for all modules using the gitreport module
    modules_config = [
        ("guidata", guidata, "."),  # guidata uses current directory
    ]
    git_repos = get_git_info_for_modules(modules_config)
    git_info_lines = format_git_info_for_pytest(git_repos, "guidata")
    if git_info_lines:
        infolist.extend(git_info_lines)

    return infolist



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)

"""
Auto-apply functionality test

This test verifies that DictItem and FloatArrayItem editors automatically
trigger the apply action when used within a DataSetEditGroupBox context.
"""

# guitest: show

import numpy as np

import guidata.dataset as gds
from guidata.dataset.qtwidgets import DataSetEditGroupBox
from guidata.env import execenv
from guidata.qthelpers import qt_app_context


class AutoApplyDataSet(gds.DataSet):
    """Test dataset with DictItem and FloatArrayItem"""

    dictionary = gds.DictItem("Dictionary", default={"a": 1, "b": 2, "c": 3})
    array = gds.FloatArrayItem("Array", default=np.array([[1, 2], [3, 4]]))
    string = gds.StringItem("String", default="test")


class AutoApplySignalChecker:
    """Helper class to track signal emissions"""

    def __init__(self, groupbox: DataSetEditGroupBox):
        self.groupbox = groupbox
        self.signal_received = False
        groupbox.SIG_APPLY_BUTTON_CLICKED.connect(self.on_signal)

    def on_signal(self):
        """Signal handler"""
        self.signal_received = True
        execenv.print("Signal received: SIG_APPLY_BUTTON_CLICKED")

    def reset(self):
        """Reset the checker state"""
        self.signal_received = False


def test_auto_apply_dictitem():
    """Test that DictItem widget has auto-apply functionality"""
    with qt_app_context():
        # Create the groupbox and signal checker
        groupbox = DataSetEditGroupBox("Test", AutoApplyDataSet)
        checker = AutoApplySignalChecker(groupbox)

        # Get the DictItem widget - widget.item is a DataItemVariable,
        # widget.item.item is the actual DataItem with the type/name info
        dict_widget = None
        for widget in groupbox.edit.widgets:
            if (
                hasattr(widget, "item")
                and hasattr(widget.item, "item")
                and isinstance(widget.item.item, gds.DictItem)
            ):
                dict_widget = widget
                break

        assert dict_widget is not None, "DictItem widget not found"

        # Verify the widget has the _trigger_auto_apply method
        assert hasattr(dict_widget, "_trigger_auto_apply"), (
            "DictItem widget missing _trigger_auto_apply method"
        )

        # Call _trigger_auto_apply to simulate what the dictionary editor does
        dict_widget._trigger_auto_apply()

        # Process events to allow deferred execution
        from qtpy.QtWidgets import QApplication

        QApplication.processEvents()

        # Verify signal was received
        assert checker.signal_received, "Signal was not emitted after auto-apply"
        execenv.print("✓ DictItem auto-apply triggered signal")

        # Verify button is disabled after processing events
        assert not groupbox.apply_button.isEnabled(), (
            "Apply button should be disabled after auto-apply"
        )
        execenv.print("✓ Apply button is disabled after auto-apply")


def test_auto_apply_floatarrayitem():
    """Test that FloatArrayItem widget has auto-apply functionality"""
    with qt_app_context():
        # Create the groupbox and signal checker
        groupbox = DataSetEditGroupBox("Test", AutoApplyDataSet)
        checker = AutoApplySignalChecker(groupbox)

        # Get the FloatArrayItem widget
        array_widget = None
        for widget in groupbox.edit.widgets:
            if (
                hasattr(widget, "item")
                and hasattr(widget.item, "item")
                and isinstance(widget.item.item, gds.FloatArrayItem)
            ):
                array_widget = widget
                break

        assert array_widget is not None, "FloatArrayItem widget not found"

        # Verify the widget has the _trigger_auto_apply method
        assert hasattr(array_widget, "_trigger_auto_apply"), (
            "FloatArrayItem widget missing _trigger_auto_apply method"
        )

        # Call _trigger_auto_apply to simulate what the array editor does
        array_widget._trigger_auto_apply()

        # Process events to allow deferred execution
        from qtpy.QtWidgets import QApplication

        QApplication.processEvents()

        # Verify signal was received
        assert checker.signal_received, "Signal was not emitted after auto-apply"
        execenv.print("✓ FloatArrayItem auto-apply triggered signal")

        # Verify button is disabled after processing events
        assert not groupbox.apply_button.isEnabled(), (
            "Apply button should be disabled after auto-apply"
        )
        execenv.print("✓ Apply button is disabled after auto-apply")


def test_auto_apply_widget_hierarchy():
    """Test auto-apply works when DataSetEditGroupBox is in widget hierarchy"""
    with qt_app_context():
        from qtpy.QtWidgets import QFrame, QStackedWidget, QTabWidget, QVBoxLayout

        # Create a complex widget hierarchy similar to DataLab's Properties panel
        tab_widget = QTabWidget()
        stacked = QStackedWidget()
        frame1 = QFrame()
        frame2 = QFrame()

        # Create the groupbox inside the hierarchy
        groupbox = DataSetEditGroupBox("Test", AutoApplyDataSet)
        checker = AutoApplySignalChecker(groupbox)

        # Build the hierarchy
        layout = QVBoxLayout()
        layout.addWidget(groupbox)
        frame2.setLayout(layout)
        frame1_layout = QVBoxLayout()
        frame1_layout.addWidget(frame2)
        frame1.setLayout(frame1_layout)
        stacked.addWidget(frame1)
        tab_widget.addTab(stacked, "Test Tab")

        # Get the widget and trigger auto-apply
        dict_widget = None
        for widget in groupbox.edit.widgets:
            if (
                hasattr(widget, "item")
                and hasattr(widget.item, "item")
                and isinstance(widget.item.item, gds.DictItem)
            ):
                dict_widget = widget
                break

        assert dict_widget is not None, "DictItem widget not found"

        # Simulate dictionary update and auto-apply
        dict_widget._trigger_auto_apply()

        # Process events
        from qtpy.QtWidgets import QApplication

        QApplication.processEvents()

        # Verify it still works even with complex hierarchy
        assert checker.signal_received, "Signal was not emitted in complex hierarchy"
        assert not groupbox.apply_button.isEnabled(), (
            "Apply button should be disabled after auto-apply"
        )
        execenv.print("✓ Auto-apply works correctly in complex widget hierarchy")


if __name__ == "__main__":
    # Run all tests
    test_auto_apply_dictitem()
    execenv.print("\n" + "=" * 80 + "\n")
    test_auto_apply_floatarrayitem()
    execenv.print("\n" + "=" * 80 + "\n")
    test_auto_apply_widget_hierarchy()
    execenv.print("\nAll tests passed!")



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)

# flake8: noqa

from .conv import (
    create_dataset_from_dict,
    create_dataset_from_func,
    restore_dataset,
    update_dataset,
    dataset_to_json,
    json_to_dataset,
)
from .dataitems import (
    BoolItem,
    ButtonItem,
    ChoiceItem,
    ColorItem,
    DateItem,
    DateTimeItem,
    DictItem,
    DirectoryItem,
    FileOpenItem,
    FileSaveItem,
    FilesOpenItem,
    FloatArrayItem,
    FloatItem,
    FontFamilyItem,
    ImageChoiceItem,
    IntItem,
    LabeledEnum,
    MultipleChoiceItem,
    StringItem,
    TextItem,
)
from .datatypes import (
    ActivableDataSet,
    AnyDataSet,
    assert_datasets_equal,
    BeginGroup,
    BeginTabGroup,
    ComputedProp,
    DataItemValidationWarning,
    DataItemValidationError,
    DataItem,
    DataItemProxy,
    DataItemVariable,
    DataSet,
    DataSetGroup,
    DataSetMeta,
    EndGroup,
    EndTabGroup,
    FormatProp,
    FuncProp,
    FuncPropMulti,
    GetAttrProp,
    GroupItem,
    ItemProperty,
    NoDefault,
    NotProp,
    Obj,
    ObjectItem,
    SeparatorItem,
    TabGroupItem,
    ValueProp,
)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)

"""
Test HDF5 DateTime Serialization
---------------------------------

Testing datetime/date serialization with metadata to avoid false positives.
"""

from __future__ import annotations

import atexit
import datetime
import os
import os.path as osp

import numpy as np

from guidata.io import HDF5Reader, HDF5Writer


class DateTimeTestObject:
    """Test object with various datetime and numeric values"""

    def __init__(self) -> None:
        # Datetime values that should be preserved
        self.dt1 = datetime.datetime(2024, 1, 15, 10, 30, 45)
        self.dt2 = datetime.datetime.now()
        self.date1 = datetime.date(2024, 6, 1)
        self.date2 = datetime.date.today()

        # Numeric values that could be mistaken for datetime
        # (these should remain as numbers)
        self.timestamp_like_float = 1500000000.0  # Within datetime range
        self.timestamp_like_int = 1500000000  # Within datetime range
        self.ordinal_like_int = 737000  # Within date ordinal range
        self.regular_float = 3.14159
        self.regular_int = 42

        # Dictionary with mixed types
        self.metadata = {
            "created": datetime.datetime(2023, 12, 25, 12, 0, 0),
            "modified": datetime.date(2024, 1, 1),
            "count": 100000,  # Could be mistaken for date ordinal
            "score": 1.5e8,  # Could be mistaken for timestamp
            "name": "Test Object",
        }

    def __eq__(self, other: object) -> bool:
        """Check equality"""
        if not isinstance(other, DateTimeTestObject):
            return False
        return (
            self.dt1 == other.dt1
            and self.dt2 == other.dt2
            and self.date1 == other.date1
            and self.date2 == other.date2
            and self.timestamp_like_float == other.timestamp_like_float
            and self.timestamp_like_int == other.timestamp_like_int
            and self.ordinal_like_int == other.ordinal_like_int
            and self.regular_float == other.regular_float
            and self.regular_int == other.regular_int
            and self.metadata == other.metadata
        )

    def serialize(self, writer: HDF5Writer) -> None:
        """Serialize to HDF5"""
        writer.write(self.dt1, "dt1")
        writer.write(self.dt2, "dt2")
        writer.write(self.date1, "date1")
        writer.write(self.date2, "date2")
        writer.write(self.timestamp_like_float, "timestamp_like_float")
        writer.write(self.timestamp_like_int, "timestamp_like_int")
        writer.write(self.ordinal_like_int, "ordinal_like_int")
        writer.write(self.regular_float, "regular_float")
        writer.write(self.regular_int, "regular_int")
        with writer.group("metadata"):
            writer.write_dict(self.metadata)

    def deserialize(self, reader: HDF5Reader) -> None:
        """Deserialize from HDF5"""
        self.dt1 = reader.read("dt1")
        self.dt2 = reader.read("dt2")
        self.date1 = reader.read("date1")
        self.date2 = reader.read("date2")
        self.timestamp_like_float = reader.read("timestamp_like_float")
        self.timestamp_like_int = reader.read("timestamp_like_int")
        self.ordinal_like_int = reader.read("ordinal_like_int")
        self.regular_float = reader.read("regular_float")
        self.regular_int = reader.read("regular_int")
        with reader.group("metadata"):
            self.metadata = reader.read_dict()


def test_h5fmt_datetime_serialization():
    """Test datetime serialization with metadata to avoid false positives"""
    path = osp.abspath("test_datetime.h5")
    atexit.register(os.unlink, path)

    # Create and serialize the object
    original = DateTimeTestObject()
    writer = HDF5Writer(path)
    original.serialize(writer)
    writer.close()

    # Deserialize the object
    loaded = DateTimeTestObject()
    reader = HDF5Reader(path)
    loaded.deserialize(reader)
    reader.close()

    # Verify datetime objects are correctly restored
    assert isinstance(loaded.dt1, datetime.datetime)
    assert loaded.dt1 == original.dt1
    assert isinstance(loaded.dt2, datetime.datetime)
    assert loaded.dt2 == original.dt2
    assert isinstance(loaded.date1, datetime.date)
    assert loaded.date1 == original.date1
    assert isinstance(loaded.date2, datetime.date)
    assert loaded.date2 == original.date2

    # Verify numeric values are NOT converted to datetime (no false positives)
    assert isinstance(loaded.timestamp_like_float, (float, np.floating))
    assert loaded.timestamp_like_float == original.timestamp_like_float
    assert isinstance(loaded.timestamp_like_int, (int, np.integer))
    assert loaded.timestamp_like_int == original.timestamp_like_int
    assert isinstance(loaded.ordinal_like_int, (int, np.integer))
    assert loaded.ordinal_like_int == original.ordinal_like_int
    assert isinstance(loaded.regular_float, (float, np.floating))
    assert loaded.regular_float == original.regular_float
    assert isinstance(loaded.regular_int, (int, np.integer))
    assert loaded.regular_int == original.regular_int

    # Verify dictionary values
    assert isinstance(loaded.metadata["created"], datetime.datetime)
    assert loaded.metadata["created"] == original.metadata["created"]
    assert isinstance(loaded.metadata["modified"], datetime.date)
    assert loaded.metadata["modified"] == original.metadata["modified"]
    assert isinstance(loaded.metadata["count"], (int, np.integer))
    assert loaded.metadata["count"] == original.metadata["count"]
    assert isinstance(loaded.metadata["score"], (float, np.floating))
    assert loaded.metadata["score"] == original.metadata["score"]
    assert isinstance(loaded.metadata["name"], str)
    assert loaded.metadata["name"] == original.metadata["name"]

    # Overall equality check
    assert loaded == original

    print("All datetime serialization tests passed!")


if __name__ == "__main__":
    test_h5fmt_datetime_serialization()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)

"""
All guidata item/group features demo
"""

# guitest: show

import atexit
import shutil
import tempfile

import numpy as np

import guidata.dataset as gds
from guidata.config import ValidationMode, get_validation_mode, set_validation_mode
from guidata.dataset.qtitemwidgets import DataSetWidget
from guidata.dataset.qtwidgets import DataSetEditLayout, DataSetShowLayout
from guidata.env import execenv
from guidata.qthelpers import qt_app_context

# Creating temporary files and registering cleanup functions
TEMPDIR = tempfile.mkdtemp(prefix="test_")
atexit.register(shutil.rmtree, TEMPDIR)
FILE_ETA = tempfile.NamedTemporaryFile(suffix=".eta", dir=TEMPDIR)
atexit.register(FILE_ETA.close)
FILE_CSV = tempfile.NamedTemporaryFile(suffix=".csv", dir=TEMPDIR)
atexit.register(FILE_CSV.close)


class SubDataSet(gds.DataSet):
    dir = gds.DirectoryItem("Directory", TEMPDIR)
    fname = gds.FileOpenItem("Single file (open)", ("csv", "eta"), FILE_CSV.name)
    fnames = gds.FilesOpenItem("Multiple files", "csv", [FILE_CSV.name])
    fname_s = gds.FileSaveItem("Single file (save)", "eta", FILE_ETA.name)


class SubDataSetWidget(DataSetWidget):
    klass = SubDataSet


class SubDataSetItem(gds.ObjectItem):
    klass = SubDataSet


DataSetEditLayout.register(SubDataSetItem, SubDataSetWidget)
DataSetShowLayout.register(SubDataSetItem, SubDataSetWidget)


class Parameters(gds.DataSet):
    """
    DataSet test
    The following text is the DataSet 'comment': <br>Plain text or
    <b>rich text<sup>2</sup></b> are both supported,
    as well as special characters (α, β, γ, δ, ...)
    """

    dict_ = gds.DictItem(
        "dict_item",
        {
            "strings_col": ["a", "b", "c"],
            "_col": [1, 2.0, 3],
            "float_col": 1.0,
        },
    )
    string = gds.StringItem("String", default="")
    string_regexp = gds.StringItem("String", regexp=r"^[a-z]+[0-9]$", default="abcd9")
    password = gds.StringItem("Password", default="", password=True)
    text = gds.TextItem("Text", default="")
    _bg = gds.BeginGroup("A sub group")
    float_slider = gds.FloatItem(
        "Float (with slider)", default=0.5, min=0, max=1, step=0.01, slider=True
    )
    fl1 = gds.FloatItem(
        "Current", default=10.0, min=1, max=30, unit="mA", help="Threshold current"
    )
    fl2 = gds.FloatItem("Float (col=1)", default=1.0, min=1, max=1).set_pos(col=1)
    fl3 = gds.FloatItem("Not checked float").set_prop("data", check_value=False)
    bool1 = gds.BoolItem("Boolean option without label")
    bool2 = gds.BoolItem("Boolean option with label", "Label").set_pos(col=1, colspan=2)
    color = gds.ColorItem("Color", default="red")
    choice1 = gds.ChoiceItem(
        "Single choice (radio)",
        [(16, "first choice"), (32, "second choice"), (64, "third choice")],
        radio=True,
    ).set_pos(col=1, colspan=2)
    choice2 = gds.ChoiceItem(
        "Single choice (combo)",
        [(16, "first choice"), (32, "second choice"), (64, "third choice")],
    ).set_pos(col=1, colspan=2)
    _eg = gds.EndGroup("A sub group")
    floatarray = gds.FloatArrayItem(
        "Float array", default=np.ones((50, 5), float), format=" %.2e "
    ).set_pos(col=1)
    floatarray2 = gds.FloatArrayItem(
        "Empty array", default=np.array([], float)
    ).set_pos(col=2)
    g0 = gds.BeginTabGroup("group")
    mchoice1 = gds.MultipleChoiceItem(
        "MC type 1", ["first choice", "second choice", "third choice"]
    ).vertical(2)
    mchoice2 = (
        gds.ImageChoiceItem(
            "MC type 2",
            [
                ("rect", "first choice", "gif.png"),
                ("ell", "second choice", "txt.png"),
                ("qcq", "third choice", "html.png"),
            ],
        )
        .set_pos(col=1)
        .set_prop("display", icon="file.png")
        .set_prop("display", size=(32, 32))
    )
    mchoice3 = gds.MultipleChoiceItem(
        "MC type 3", [str(i) for i in range(10)]
    ).horizontal(2)
    eg0 = gds.EndTabGroup("group")
    integer_slider = gds.IntItem(
        "Integer (with slider)", default=5, min=-50, max=50, slider=True
    )
    integer = gds.IntItem("Integer", default=5, min=3, max=60).set_pos(col=1)


def test_all_features():
    """Test all guidata item/group features"""
    old_mode = get_validation_mode()
    set_validation_mode(ValidationMode.STRICT)
    with execenv.context(accept_dialogs=True):
        with qt_app_context():
            prm1 = Parameters()
            prm1.floatarray[:, 0] = np.linspace(-5, 5, 50)
            execenv.print(prm1)

            if prm1.edit():
                prm1.edit()
                execenv.print(prm1)
            prm1.view()

            prm2 = Parameters.create(integer=59, string="Using `create`")
            assert prm2.integer == 59
            print(prm2)

            try:
                # Try to set an unknown attribute using the `create` method:
                Parameters.create(unknown_attribute=42)
            except AttributeError:
                pass
            else:
                raise AssertionError("AttributeError not raised")

            execenv.print("OK")
    set_validation_mode(old_mode)


if __name__ == "__main__":
    test_all_features()



# -*- coding: utf-8 -*-

import atexit
import datetime
import shutil
import tempfile

import guidata.dataset as gds

# Creating temporary files and registering cleanup functions
TEMPDIR = tempfile.mkdtemp(prefix="test_")
atexit.register(shutil.rmtree, TEMPDIR)
FILE_ETA = tempfile.NamedTemporaryFile(suffix=".eta", dir=TEMPDIR)
atexit.register(FILE_ETA.close)
FILE_CSV = tempfile.NamedTemporaryFile(suffix=".csv", dir=TEMPDIR)
atexit.register(FILE_CSV.close)


class AutodocExampleParam1(gds.DataSet):
    """Example of a complete dataset with all possible items. Used as an autodoc
    example."""

    dir = gds.DirectoryItem("<strong>Directory</strong>", TEMPDIR)
    a = gds.FloatItem("Parameter #1", default=2.3)
    b = gds.IntItem("Parameter #2", min=0, max=10, default=5)
    c = gds.StringItem("Parameter #3", default="default value")
    type = gds.ChoiceItem("Processing algorithm", ("type 1", "type 2", "type 3"))
    fname = gds.FileOpenItem("Open file", ("csv", "eta"), FILE_CSV.name)
    fnames = gds.FilesOpenItem("Open files", "csv", FILE_CSV.name)
    fname_s = gds.FileSaveItem("Save file", "eta", FILE_ETA.name)
    string = gds.StringItem("String")
    text = gds.TextItem("Text")
    float_slider = gds.FloatItem(
        "Float (with slider)", default=0.5, min=0, max=1, step=0.01, slider=True
    )
    integer = gds.IntItem("Integer", default=5, min=3, max=16, slider=True).set_pos(
        col=1
    )
    dtime = gds.DateTimeItem("Date/time", default=datetime.datetime(2010, 10, 10))
    date = gds.DateItem("Date", default=datetime.date(2010, 10, 10)).set_pos(col=1)
    bool1 = gds.BoolItem("Boolean option without label")
    bool2 = gds.BoolItem("Boolean option with label", "Label")
    _bg = gds.BeginGroup("A sub group")
    color = gds.ColorItem("Color", default="red")
    choice = gds.ChoiceItem(
        "Single choice 1",
        [
            ("16", "first choice"),
            ("32", "second choice"),
            ("64", "third choice"),
            (128, "fourth choice"),
        ],
    )
    mchoice2 = gds.ImageChoiceItem(
        "Single choice 2",
        [
            ("rect", "first choice", "gif.png"),
            ("ell", "second choice", "txt.png"),
            ("qcq", "third choice", "file.png"),
        ],
    )
    _eg = gds.EndGroup("A sub group")
    floatarray = gds.FloatArrayItem("Float array", format=" %.2e ").set_pos(col=1)
    mchoice3 = gds.MultipleChoiceItem(
        "MC type 1", [str(i) for i in range(12)]
    ).horizontal(4)
    mchoice1 = (
        gds.MultipleChoiceItem(
            "MC type 2", ["first choice", "second choice", "third choice"]
        )
        .vertical(1)
        .set_pos(col=1)
    )
    dictionary = gds.DictItem(
        "Dictionary",
        help="This is a dictionary",
    )

    def doc_test(self, a: int, b: float, c: str) -> str:
        """Test method for autodoc.

        Args:
            a: first parameter.
            b: second parameter.
            c: third parameter.

        Returns:
            Concatenation of c and (a + b).
        """
        return c + str(a + b)


class AutodocExampleParam2(AutodocExampleParam1):
    pass



# -*- coding: utf-8 -*-
"""
guidata
=======

Based on the Qt library :mod:`guidata` is a Python library generating graphical
user interfaces for easy dataset editing and display. It also provides helpers
and application development tools for Qt.
"""

__version__ = "3.13.3"


# Dear (Debian, RPM, ...) package makers, please feel free to customize the
# following path to module's data (images) and translations:
DATAPATH = LOCALEPATH = ""


import guidata.config  # noqa: E402, F401


def qapplication():
    """
    Return QApplication instance
    Creates it if it doesn't already exist
    """
    from qtpy.QtWidgets import QApplication

    app = QApplication.instance()
    if not app:
        app = QApplication([])
        install_translator(app)
        from guidata import qthelpers

        qthelpers.set_color_mode()
    return app


QT_TRANSLATOR = None


def install_translator(qapp):
    """Install Qt translator to the QApplication instance"""
    global QT_TRANSLATOR
    if QT_TRANSLATOR is None:
        from qtpy.QtCore import QLibraryInfo, QLocale, QTranslator

        locale = QLocale.system().name()
        # Qt-specific translator
        qt_translator = QTranslator()
        paths = QLibraryInfo.location(QLibraryInfo.TranslationsPath)
        for prefix in ("qt", "qtbase"):
            if qt_translator.load(prefix + "_" + locale, paths):
                QT_TRANSLATOR = qt_translator  # Keep reference alive
                break
    if QT_TRANSLATOR is not None:
        qapp.installTranslator(QT_TRANSLATOR)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)
#
# The array editor subpackage was derived from Spyder's arrayeditor.py module
# which is licensed under the terms of the MIT License (see spyder/__init__.py
# for details), copyright © Spyder Project Contributors

# pylint: disable=C0103
# pylint: disable=R0903
# pylint: disable=R0911
# pylint: disable=R0201

"""
Provide classes to wrap Numpy arrays and handle changes made to them. Array handlers
acts as a pointer to the original array and allows to share the same array in multiple
models/widgets and views. They handle data access and changes.
"""

from __future__ import annotations

import copy
from typing import Any, Generic, TypeVar, Union, cast

import numpy as np

# TODO: (When Python 3.8-3.9 support is dropped) Use `np.ndarray | np.ma.MaskedArray`
# instead of `Union[np.ndarray, np.ma.MaskedArray]`
AnySupportedArray = TypeVar(
    "AnySupportedArray", bound=Union[np.ndarray, np.ma.MaskedArray]
)
AnyArrayHandler = TypeVar("AnyArrayHandler", bound="BaseArrayHandler")


class BaseArrayHandler(Generic[AnySupportedArray]):
    """Wrapper class around a Numpy nparray that is used a pointer to share the same
    array in multiple models/widgets and views. It handles data access and changes.

    Args:
        array: Numpy array to wrap
        variable_size: Flag to indicate if the size of the array can be modified (i.e.
         that data can be inserted on any axis). This flag changes the underlying
         data management strategy. If True, the original array will be copied and
         changes will be done inplace in the copy, else changes are stored in a
         dictionnary and applied inplace when the appropriate method is called.
         Defaults to False.
    """

    __slots__ = (
        "_variable_size",
        "_backup_array",
        "_array",
        "_dtype",
        "current_changes",
        "_og_shape",
    )

    def __init__(
        self,
        array: AnySupportedArray,
        variable_size: bool = False,
    ) -> None:
        self._variable_size = variable_size
        self._og_shape = None
        self._array = array
        self._backup_array = array
        self._init_arrays(array)

        self._dtype = array.dtype
        self.current_changes: dict[tuple[str | int, ...] | str, bool] = {}

    def _init_arrays(self, array: np.ndarray | np.ma.MaskedArray) -> None:
        """Small method to handle variable initializations dependent on the array.

        Args:
            array: Numpy array to use
        """
        if self._variable_size:
            self._backup_array = array
            if array.ndim == 1:
                self._array = array.reshape(-1, 1)
            else:
                self._array = copy.deepcopy(array)

        else:
            if array.ndim == 1:
                self._og_shape = array.shape
                array.shape = (array.size, 1)
            self._array = array

    @property
    def variable_size(self) -> bool:
        """Returns the variable_size flag. If True, the array is resizable and changes

        Returns
            Variable size flag
        """
        return self._variable_size

    @property
    def ndim(self) -> int:
        """Numpy ndim property.

        Returns
            Number of dimensions of the array
        """
        return self._array.ndim

    @property
    def flags(self) -> np.flagsobj:
        """Numpy flags property.

        Returns
            Flags of the array"""
        return self._array.flags

    @property
    def shape(self) -> tuple[int, ...]:
        """Numpy shape property.

        Returns
            Shape of the array
        """
        return self._array.shape

    @shape.setter
    def shape(self, value: tuple[int, ...]) -> None:
        """Numpy shape property setter. Sets the shape of the array.

        Args:
            value: new shape of the array
        """
        self._array.shape = value

    @property
    def dtype(self) -> Any:
        """Numpy dtype property.

        Returns
            dtype of the array
        """
        return self._dtype

    @property
    def data(self) -> memoryview:
        """Used to get the underlying data of an array. Useful for Numpy's structured
        arrays.

        Returns
            A memoryview to the underlying data (= a np.ndarray)
        """
        return self._array.data

    def insert_on_axis(
        self, index: int, axis: int, insert_number: int = 1, default: Any = 0
    ) -> None:
        """Insert new row(s) of default values on an axis. Makes a copy of the original
        array.

        Args:
            index: Index from which to start the insertion.
            axis: axis on which to make the insertion.
            insert_number: Number of rows to insert at once. Defaults to 1.
            default: default value to insert. Defaults to the "zero value" of a type
             (e.g. "" for str or False for Booleans). Defaults to 0.
        """
        indexes = (index,) * insert_number
        self._array = np.insert(self._array, indexes, default, axis=axis)

    def delete_on_axis(self, index: int, axis: int, remove_number: int = 1):
        """Delete row(s) on an axis. Makes a copy of the original array.

        Args:
            index: index from which to start the deletion.
            axis: axis on which to make the deletion.
            remove_number: Number of rows to delete at once. Defaults to 1.
        """
        indexes = range(index, index + remove_number)
        self._array = np.delete(self._array, indexes, axis=axis)

    def get_array(self) -> AnySupportedArray:
        """Returns the current wrapped array. If variable_size is False, the returned
        array does not contain the current modifications as they are saved separately.

        Returns
            Numpy array contained in the handler instance
        """
        return cast(AnySupportedArray, self._array)

    def new_row(self, index: int, insert_number: int = 1, default: Any = 0):
        """Insert new row(s) on axis 0. Do not for 3D+ arrays. Prefer the method
        'insert_on_axis' instead to be sure to insert on the right axis from the Model.

        Args:
            index: Index from which to start the insertion.
            insert_number: Number of rows to insert at once. Defaults to 1.
            default: default value to insert. Defaults to the "zero value" of a type
             (e.g. "" for str or False for Booleans). Defaults to 0.
        """
        self.insert_on_axis(index, 0, insert_number, default)

    def new_col(self, index: int, insert_number: int = 1, default: Any = 0):
        """Insert new column(s) on axis 1. Do not for 3D+ arrays. Prefer the method
        'insert_on_axis' instead to be sure to insert on the right axis from the Model.

        Args:
            index: Index from which to start the insertion.
            insert_number: Number of columns to insert at once. Defaults to 1.
            default: default value to insert. Defaults to the "zero value" of a type
             (e.g. "" for str or False for Booleans). Defaults to 0.
        """
        self.insert_on_axis(index, 1, insert_number, default)

    def __setitem__(self, key: tuple[int, ...] | str, item: Any):
        """Data setter that allows to use this class like the original array. The data
        storage strategy changes depending on the variable_size flag.

        Args:
            key: key to set (=array index)
            item: value to set. Must be of the same type as the array data.
        """
        if self._variable_size:
            self._array[key] = item
            return
        self.current_changes[key] = item

    def __getitem__(self, key: tuple[int, ...] | str) -> Any:
        """Data getter that allows to use this class like the original array. The data
        retrieval strategy changes depending on the variable_size flag.

        Args:
            key: key to get (=array index)

        Returns:
            The requested value from the array
        """
        if not self._variable_size:
            return self.current_changes.get(key, self._array[key])
        return self._array[key]

    def apply_changes(self) -> None:
        """Apply changes. Only useful if flag viariable_size is False as it will write
        the values stored in a dictionary in the original array. This operation is
        non-reversible as it writes inplace.
        """
        if not self._variable_size:
            for coor, value in self.current_changes.items():
                self._array[coor] = value
            self.current_changes.clear()
        else:
            self._backup_array = copy.deepcopy(self._array)

    def clear_changes(self) -> None:
        """Deletes all the changes made until that point. If the variable_size flag is
        True, then restores a copy of the origninal array (makes new copy needed). Else,
        clears the dictionnary where changes are temporarily stored.
        """
        if not self._variable_size:
            self.current_changes.clear()
        else:
            self._init_arrays(self._backup_array)

    def reset_shape_if_changed(self) -> None:
        """When a numpy array is 1D, the handler changes the shape to add a second
        dimension of size 1 to act like a normal 2d array in the BaseArrayModel. In some
        instances, the shape must be reset (i.e. when getting the array when the
        ArrayEditor is closed). When the shape is reset, the array editor may not work
        properly as the awaited 2nd dimension is removed. The method _init_arrays(array)
        or clear_changes() must be called to avoid errors possible IndexError.
        """
        if self._og_shape is not None:
            self._array.shape = self._og_shape


class MaskedArrayHandler(BaseArrayHandler[np.ma.MaskedArray]):
    """Same as the class BaseArrayHandler but with additionnal functionnalities to
    handled a Numpy MaskedArray.

    Args:
        array: Numpy MaskedArray to wrap
        variable_size: array resizability flag, refer to BaseArrayHandler for more
         information. Defaults to False.
    """

    __slots__ = ("current_mask_changes",)
    _array: np.ma.MaskedArray
    _backup_array: np.ma.MaskedArray

    def __init__(
        self,
        array: np.ma.MaskedArray,
        variable_size: bool = False,
    ) -> None:
        super().__init__(array, variable_size)
        self.current_mask_changes: dict[tuple[int, ...], Any] = {}

    @property
    def mask(self) -> np.ndarray:
        """Numpy mask property.

        Returns:
            The mask of the array
        """
        return self._array.mask

    def insert_on_axis(
        self,
        index: int,
        axis: int,
        insert_number: int = 1,
        default: Any = 0,
        default_mask: bool = False,
    ) -> None:
        """Refer to BaseArrayHandler.insert_on_axis for the full details.
        The only difference is that with this method is that it also inserts default
        values for the mask.

        Args:
            index: index from which to start the insertion.
            axis: axis on which to make the insertion.
            insert_number: Number of rows to insert at once. Defaults to 1.
            default: default value to insert. Defaults to the "zero value" of a type
             (e.g. "" for str or False for Booleans). Defaults to 0.
            default_mask: default mask value to insert. Defaults to False.
        """
        indexes = (index,) * insert_number
        new_array: np.ma.MaskedArray = np.insert(
            self._array, indexes, default, axis=axis
        )
        # The check is performed at init and array type cannot change
        new_mask = self._array.mask
        new_mask = np.insert(new_mask, indexes, default_mask, axis=axis)
        new_array.mask = new_mask
        self._array = new_array

    def delete_on_axis(self, index: int, axis: int, remove_number: int = 1) -> None:
        """Refer to BaseArrayHandler.delete_on_axis for the full details.
        The only difference is that with this method is that it also keep the previous
        value of the mask.

        Args:
            index: index from which to start the deletion.
            axis: axis on which to make the deletion.
            remove_number: number of rows to delete. Defaults to 1.
        """
        # indexes = (index,) * remove_number
        indexes = range(index, min(index + remove_number, self._array.shape[axis]))
        new_array: np.ma.MaskedArray = np.delete(self._array, indexes, axis=axis)
        # The check is performed at init and array type cannot change
        new_mask = self._array.mask
        new_mask = np.delete(new_mask, indexes, axis=axis)
        new_array.mask = new_mask
        self._array = new_array

    def set_mask_value(self, key: tuple[int, ...], value: bool) -> None:
        """Setter for the mask values. Identical to BaseArrayHandler.__setitem__ but for
        the mask.

        Args:
            key:  key to set (=mask index).
            item: value to set.
        """
        if not self._variable_size:
            self.current_mask_changes[key] = value
        else:
            self._array.mask[key] = value

    def get_mask_value(self, key: tuple[int, ...]) -> bool:
        """Getter for the mask values. Identical to BaseArrayHandler.__getitem__ but for
        the mask.

        Args:
            key:  key to get (=mask index).

        Returns:
            The requested value from the mask
        """
        if not self._variable_size:
            return self.current_mask_changes.get(key, self._array.mask[key])
        return self._array.mask[key]

    def get_data_value(self, key: tuple[int, ...]) -> Any:
        """Setter for the data values (unmasked). Identical to
        BaseArrayHandler.__setitem__ but for the unmasked array data.

        Args:
            key:  key to set (=mask index). Must be the same type as the array.
            item: value to set.

        Returns:
            The requested value from the array
        """
        if not self._variable_size:
            return self.current_changes.get(key, self._array.data[key])
        return self._array.data[key]

    def set_data_value(self, key: tuple[int, ...], value: bool) -> None:
        """Getter for the data values (unmasked). Identical to
        BaseArrayHandler.__getitem__ but for the unmasked array data.

        Args:
            key:  key to get (=array index).
        """
        if not self._variable_size:
            self.current_changes[key] = value
        else:
            self._array.data[key] = value

    def apply_changes(self) -> None:
        """Same as BaseArrayHandler.apply_changes but also applies changes to the
        mask."""
        super().apply_changes()
        for coor, value in self.current_mask_changes.items():
            self._array.mask[coor] = value
        self.current_mask_changes.clear()

    def clear_changes(self) -> None:
        """Same as BaseArrayHandler.clear_changes but also clears the changes made to
        the mask."""
        super().clear_changes()
        if not self._variable_size:
            self.current_mask_changes.clear()


class RecordArrayHandler(BaseArrayHandler[np.ndarray]):
    """Same as the class BaseArrayHandler but with additionnal functionnalities to
    handled Numpy's structured arrays.

    Args:
        array: Numpy ndarray
        variable_size: array resizability flag, refer to BaseArrayHandler for more
         information. Defaults to False.
    """

    def __init__(
        self,
        array: np.recarray,
        variable_size: bool = False,
    ) -> None:
        super().__init__(array, variable_size)

    def get_record_value(self, name: str, key: tuple[str | int, ...]) -> Any:
        """Getter for the Numpy's structured array. Identical to
        BaseArrayHandler.__getitem__ but for the named values.

        Args:
            name: type name to get.
            key:  key to get.

        Returns:
            The requested value from the array
        """
        if not self._variable_size:
            return self.current_changes.get((name, *key), self._array[name][key])
        return self._array[name][key]

    def set_record_value(
        self, name: str, key: tuple[str | int, ...], value: Any
    ) -> None:
        """Setter for the Numpy's structured array. Identical to
        BaseArrayHandler.__setitem__ but for the named values.

        Args:
            name: type name to get.
            key:  key to set (i.e. array index).
        """
        if not self._variable_size:
            self.current_changes[(name, *key)] = value
        else:
            self._array[name][key] = value



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License
"""
Tests for objecteditor.py
"""

# guitest: show

import datetime

import numpy as np

try:
    from PIL import Image
except ImportError:
    # PIL is not installed
    Image = None

from guidata.env import execenv
from guidata.qthelpers import qt_app_context
from guidata.widgets.objecteditor import oedit


def test_objecteditor():
    """Run object editor test"""
    with qt_app_context():
        example = {
            "str": "kjkj kj k j j kj k jkj",
            "list": [1, 3, 4, "kjkj", None],
            "dict": {"d": 1, "a": np.random.rand(10, 10), "b": [1, 2]},
            "float": 1.2233,
            "array": np.random.rand(10, 10),
            "date": datetime.date(1945, 5, 8),
            "datetime": datetime.datetime(1945, 5, 8),
        }
        if Image is not None:
            data = np.random.randint(255, size=(100, 100)).astype("uint8")
            image = Image.fromarray(data)
            example["image"] = image
            image = oedit(image)

        class Foobar:
            """ """

            def __init__(self):
                self.text = "toto"

        foobar = Foobar()
        execenv.print(oedit(foobar))
        execenv.print(oedit(example))
        execenv.print(oedit(np.random.rand(10, 10)))
        execenv.print(oedit(oedit.__doc__))
        execenv.print(example)
        execenv.print("OK")


if __name__ == "__main__":
    test_objecteditor()



# -*- coding: utf-8 -*-

import os
import sys

sys.path.insert(0, os.path.abspath(".."))
sys.path.insert(0, os.path.abspath("autodoc"))

import guidata  # noqa: E402

creator = "Pierre Raybaut"
project = "guidata"
copyright = "2009 CEA, " + creator
version = ".".join(guidata.__version__.split(".")[:2])
release = guidata.__version__

extensions = [
    "sphinx.ext.autodoc",
    "myst_parser",
    "sphinx.ext.intersphinx",
    "sphinx.ext.doctest",
    "sphinx_copybutton",
    "sphinx.ext.napoleon",
    "sphinx_qt_documentation",
    "guidata.dataset.autodoc",
]
if "htmlhelp" in sys.argv:
    extensions += ["sphinx.ext.imgmath"]
else:
    extensions += ["sphinx.ext.mathjax"]
templates_path = ["_templates"]
source_suffix = ".rst"
master_doc = "index"

exclude_trees = []
pygments_style = "sphinx"
modindex_common_prefix = ["guidata."]
autodoc_member_order = "bysource"

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "h5py": ("https://docs.h5py.org/en/stable/", None),
}
# nitpicky = True  # Uncomment to warn about all broken links

if "htmlhelp" in sys.argv:
    html_theme = "classic"
else:
    html_theme = "python_docs_theme"
html_title = "%s %s Manual" % (project, version)
html_short_title = "%s Manual" % project
html_logo = "images/guidata-vertical.png"
html_favicon = "_static/favicon.ico"
html_static_path = ["_static"]
html_use_modindex = True
htmlhelp_basename = "guidata"
latex_documents = [
    ("index", "guidata.tex", "guidata Manual", creator, "manual"),
]



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License


"""
Tests for codeeditor.py
"""

# guitest: show

from guidata.configtools import get_icon
from guidata.env import execenv
from guidata.qthelpers import qt_app_context
from guidata.widgets.console import Console


def test_console():
    """Test Console widget."""
    with qt_app_context(exec_loop=True):
        widget = Console(debug=False, multithreaded=True)
        widget.resize(800, 600)
        widget.setWindowTitle("Console")
        widget.setWindowIcon(get_icon("guidata.svg"))
        widget.show()
        execenv.print("OK")


if __name__ == "__main__":
    test_console()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)

"""
Demonstrates how items may trigger callbacks when activated, or how
the list of choices may be dynamically changed.
"""

# guitest: show

import guidata.dataset as gds
from guidata.env import execenv
from guidata.qthelpers import qt_app_context


class Parameters(gds.DataSet):
    """Example dataset"""

    def cb_example(self, item, value):
        execenv.print("\nitem: ", item, "\nvalue:", value)
        if self.results is None:
            self.results = ""
        self.results += str(value) + "\n"
        execenv.print("results:", self.results)

    def update_x1plusx2(self, item, value):
        execenv.print("\nitem: ", item, "\nvalue:", value)
        if self.x1 is not None and self.x2 is not None:
            self.x1plusx2 = self.x1 + self.x2
        else:
            self.x1plusx2 = None

    string = gds.StringItem("String", default="foobar").set_prop(
        "display", callback=cb_example
    )
    grp1 = gds.BeginGroup("Group 1")
    x1 = gds.FloatItem("x1").set_prop("display", callback=update_x1plusx2)
    x2 = gds.FloatItem("x2").set_prop("display", callback=update_x1plusx2)
    _grp1 = gds.EndGroup("Group 1")
    grp2 = gds.BeginGroup("Group 2")
    x1plusx2 = gds.FloatItem("x1+x2").set_prop("display", active=False)
    _grp2 = gds.EndGroup("Group 2")
    boolean = gds.BoolItem("Boolean", default=True).set_prop(
        "display", callback=cb_example
    )
    color = gds.ColorItem("Color", default="red").set_prop(
        "display", callback=cb_example
    )

    def choices_callback(self, item, value):
        """Choices callback: this demonstrates how to dynamically change
        the list of choices... even if it is not very useful in this case

        Note that `None` is systematically added as the third element of
        the returned tuples: that is to ensure the compatibility between
        `ChoiceItem` and `ImageChoiceItem` (see `guidata.dataset.dataitems`)
        """
        execenv.print(f"[choices_callback]: item={item}, value={value}")
        return [
            (16, "first choice", None),
            (32, "second choice", None),
            (64, "third choice", None),
        ]

    choice = (
        gds.ChoiceItem(
            "Single choice",
            choices_callback,
            default=64,
        )
        .set_pos(col=1, colspan=2)
        .set_prop("display", callback=cb_example)
    )
    results = gds.TextItem("Results").set_prop("display", callback=cb_example)


def test_callbacks():
    """Test callbacks"""
    with qt_app_context():
        prm = Parameters()
        execenv.print(prm)
        if prm.edit():
            execenv.print(prm)
        prm.view()
        execenv.print("OK")


if __name__ == "__main__":
    test_callbacks()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)

"""
Demonstrates how items may be computed based on other items' values.

The computed items feature allows you to create read-only items whose values are
automatically calculated using a method of the dataset. This is similar to computed
properties in other frameworks.
"""

# guitest: show

import guidata.dataset as gds
from guidata.env import execenv
from guidata.qthelpers import qt_app_context


class Parameters(gds.DataSet):
    """Example dataset with computed items"""

    def compute_sum(self) -> float:
        """Compute the sum of x1 and x2"""
        return self.x1 + self.x2

    def compute_results(self) -> str:
        """Compute a results string based on current values"""
        sum_value = self.x1 + self.x2
        return f"""Results:

    String: {self.string}
    Sum: {sum_value:.2f}
    Boolean: {self.boolean}
    Color: {self.color}
    Choice: {self.choice}"""

    string = gds.StringItem("String", default="foobar")
    x1 = gds.FloatItem("x1", default=2.0)
    x2 = gds.FloatItem("x2", default=3.0)
    # This item will be computed by calling the compute_sum method
    x1plusx2 = gds.FloatItem("x1+x2 (computed)").set_computed(compute_sum)
    boolean = gds.BoolItem("Boolean", default=True)
    color = gds.ColorItem("Color", default="red")
    choice = gds.ChoiceItem(
        "Single choice",
        (("First", "first"), ("Second", "second"), ("Third", "third")),
        default="first",
    ).set_pos(col=1, colspan=2)
    # This item will be computed by calling the compute_results method
    results = gds.TextItem("Results (computed)").set_computed(compute_results)


def test_computed_items():
    """Test computed items"""
    with qt_app_context():
        prm = Parameters()

        # Test that computed items work
        execenv.print(f"x1: {prm.x1}")
        execenv.print(f"x2: {prm.x2}")
        execenv.print(f"x1+x2 (computed): {prm.x1plusx2}")
        execenv.print(f"Results (computed):\n{prm.results}")

        # Test that computed items are read-only
        try:
            prm.x1plusx2 = 10.0
            raise Exception("Should not be able to set computed item!")
        except ValueError as e:
            execenv.print(f"✓ Correctly prevented setting computed item: {e}")

        # Test that computed items update when dependencies change
        prm.x1 = 5.0
        prm.x2 = 7.0
        execenv.print("\nAfter changing x1=5.0, x2=7.0:")
        execenv.print(f"x1+x2 (computed): {prm.x1plusx2}")
        execenv.print(f"Results (computed):\n{prm.results}")

        execenv.print("\n" + "=" * 60)
        execenv.print("REAL-TIME GUI UPDATE TEST:")
        execenv.print("When you edit values in the GUI, computed items should")
        execenv.print("update automatically in real-time as you type!")
        execenv.print("Try changing x1 and x2 values and watch the computed")
        execenv.print("fields update automatically.")
        execenv.print("=" * 60)

        execenv.print(prm)
        if prm.edit():
            execenv.print(prm)
        prm.view()
        execenv.print("OK")


if __name__ == "__main__":
    test_computed_items()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)

"""
guidata test package
====================
"""

import os.path as osp

import guidata
from guidata.configtools import get_module_data_path

TESTDATAPATH = get_module_data_path("guidata", osp.join("tests", "data"))


def get_path(filename: str) -> str:
    """Return absolute path of test file"""
    return osp.join(TESTDATAPATH, filename)


def run():
    """Run guidata test launcher"""
    from guidata.guitest import run_testlauncher

    run_testlauncher(guidata)


if __name__ == "__main__":
    run()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see guidata/LICENSE for details)

"""
Handle *guidata* module configuration
(options, images and icons)
"""

import contextlib
import enum
import os.path as osp
from typing import Generator

from guidata.configtools import add_image_module_path, get_translation
from guidata.userconfig import UserConfig

APP_NAME = "guidata"
APP_PATH = osp.dirname(__file__)
add_image_module_path("guidata", "data/icons")
_ = get_translation("guidata")


def gen_mono_font_settings(size, other_settings=None):
    """Generate mono font settings"""
    settings = dict({} if other_settings is None else other_settings)
    settings.update(
        {
            "font/family/nt": ["Cascadia Code", "Consolas", "Courier New"],
            "font/family/posix": "Bitstream Vera Sans Mono",
            "font/family/mac": "Monaco",
            "font/size": size,
        }
    )
    return settings


def get_old_log_fname(fname):
    """Return old log fname from current log fname"""
    return osp.splitext(fname)[0] + ".1.log"


DEFAULTS = {
    "faulthandler": {"enabled": False, "log_path": f".{APP_NAME}_faulthandler.log"},
    "arrayeditor": gen_mono_font_settings(9),
    "dicteditor": gen_mono_font_settings(9),
    "texteditor": gen_mono_font_settings(9),
    "codeeditor": gen_mono_font_settings(10),
    "console": gen_mono_font_settings(
        9,
        {
            "cursor/width": 2,
            "codecompletion/size": (300, 180),
            "codecompletion/case_sensitive": True,
            "external_editor/path": "SciTE",
            "external_editor/gotoline": "-goto:",
        },
    ),
    "color_schemes": {
        "names": [
            "emacs",
            "idle",
            "monokai",
            "pydev",
            "scintilla",
            "spyder",
            "spyder/dark",
            "zenburn",
            "solarized/light",
            "solarized/dark",
        ],
        "default/light": "spyder",
        "default/dark": "spyder/dark",
        # ---- Emacs ----
        "emacs/name": "Emacs",
        #      Name            Color     Bold  Italic
        "emacs/background": "#000000",
        "emacs/currentline": "#2b2b43",
        "emacs/currentcell": "#1c1c2d",
        "emacs/occurrence": "#abab67",
        "emacs/ctrlclick": "#0000ff",
        "emacs/sideareas": "#555555",
        "emacs/matched_p": "#009800",
        "emacs/unmatched_p": "#c80000",
        "emacs/normal": ("#ffffff", False, False),
        "emacs/keyword": ("#3c51e8", False, False),
        "emacs/builtin": ("#900090", False, False),
        "emacs/definition": ("#ff8040", True, False),
        "emacs/comment": ("#005100", False, False),
        "emacs/string": ("#00aa00", False, True),
        "emacs/number": ("#800000", False, False),
        "emacs/instance": ("#ffffff", False, True),
        # ---- IDLE ----
        "idle/name": "IDLE",
        #      Name            Color     Bold  Italic
        "idle/background": "#ffffff",
        "idle/currentline": "#f2e6f3",
        "idle/currentcell": "#feefff",
        "idle/occurrence": "#e8f2fe",
        "idle/ctrlclick": "#0000ff",
        "idle/sideareas": "#efefef",
        "idle/matched_p": "#99ff99",
        "idle/unmatched_p": "#ff9999",
        "idle/normal": ("#000000", False, False),
        "idle/keyword": ("#ff7700", True, False),
        "idle/builtin": ("#900090", False, False),
        "idle/definition": ("#0000ff", False, False),
        "idle/comment": ("#dd0000", False, True),
        "idle/string": ("#00aa00", False, False),
        "idle/number": ("#924900", False, False),
        "idle/instance": ("#777777", True, True),
        # ---- Monokai ----
        "monokai/name": "Monokai",
        #      Name              Color     Bold  Italic
        "monokai/background": "#1f1f1f",
        "monokai/currentline": "#484848",
        "monokai/currentcell": "#3d3d3d",
        "monokai/occurrence": "#666666",
        "monokai/ctrlclick": "#0000ff",
        "monokai/sideareas": "#2a2b24",
        "monokai/matched_p": "#688060",
        "monokai/unmatched_p": "#bd6e76",
        "monokai/normal": ("#ddddda", False, False),
        "monokai/keyword": ("#f92672", False, False),
        "monokai/builtin": ("#ae81ff", False, False),
        "monokai/definition": ("#a6e22e", False, False),
        "monokai/comment": ("#75715e", False, True),
        "monokai/string": ("#e6db74", False, False),
        "monokai/number": ("#ae81ff", False, False),
        "monokai/instance": ("#ddddda", False, True),
        # ---- Pydev ----
        "pydev/name": "Pydev",
        #      Name            Color     Bold  Italic
        "pydev/background": "#ffffff",
        "pydev/currentline": "#e8f2fe",
        "pydev/currentcell": "#eff8fe",
        "pydev/occurrence": "#ffff99",
        "pydev/ctrlclick": "#0000ff",
        "pydev/sideareas": "#efefef",
        "pydev/matched_p": "#99ff99",
        "pydev/unmatched_p": "#ff99992",
        "pydev/normal": ("#000000", False, False),
        "pydev/keyword": ("#0000ff", False, False),
        "pydev/builtin": ("#900090", False, False),
        "pydev/definition": ("#000000", True, False),
        "pydev/comment": ("#c0c0c0", False, False),
        "pydev/string": ("#00aa00", False, True),
        "pydev/number": ("#800000", False, False),
        "pydev/instance": ("#000000", False, True),
        # ---- Scintilla ----
        "scintilla/name": "Scintilla",
        #         Name             Color     Bold  Italic
        "scintilla/background": "#ffffff",
        "scintilla/currentline": "#e1f0d1",
        "scintilla/currentcell": "#edfcdc",
        "scintilla/occurrence": "#ffff99",
        "scintilla/ctrlclick": "#0000ff",
        "scintilla/sideareas": "#efefef",
        "scintilla/matched_p": "#99ff99",
        "scintilla/unmatched_p": "#ff9999",
        "scintilla/normal": ("#000000", False, False),
        "scintilla/keyword": ("#00007f", True, False),
        "scintilla/builtin": ("#000000", False, False),
        "scintilla/definition": ("#007f7f", True, False),
        "scintilla/comment": ("#007f00", False, False),
        "scintilla/string": ("#7f007f", False, False),
        "scintilla/number": ("#007f7f", False, False),
        "scintilla/instance": ("#000000", False, True),
        # ---- Spyder ----
        "spyder/name": "Spyder",
        #       Name            Color     Bold  Italic
        "spyder/background": "#ffffff",
        "spyder/currentline": "#f7ecf8",
        "spyder/currentcell": "#fdfdde",
        "spyder/occurrence": "#ffff99",
        "spyder/ctrlclick": "#0000ff",
        "spyder/sideareas": "#efefef",
        "spyder/matched_p": "#99ff99",
        "spyder/unmatched_p": "#ff9999",
        "spyder/normal": ("#000000", False, False),
        "spyder/keyword": ("#0000ff", False, False),
        "spyder/builtin": ("#900090", False, False),
        "spyder/definition": ("#000000", True, False),
        "spyder/comment": ("#adadad", False, True),
        "spyder/string": ("#00aa00", False, False),
        "spyder/number": ("#800000", False, False),
        "spyder/instance": ("#924900", False, True),
        # ---- Spyder/Dark ----
        "spyder/dark/name": "Spyder Dark",
        #           Name             Color     Bold  Italic
        "spyder/dark/background": "#1f1f1f",
        "spyder/dark/currentline": "#2b2b43",
        "spyder/dark/currentcell": "#31314e",
        "spyder/dark/occurrence": "#abab67",
        "spyder/dark/ctrlclick": "#0000ff",
        "spyder/dark/sideareas": "#282828",
        "spyder/dark/matched_p": "#009800",
        "spyder/dark/unmatched_p": "#c80000",
        "spyder/dark/normal": ("#ffffff", False, False),
        "spyder/dark/keyword": ("#558eff", False, False),
        "spyder/dark/builtin": ("#aa00aa", False, False),
        "spyder/dark/definition": ("#ffffff", True, False),
        "spyder/dark/comment": ("#7f7f7f", False, False),
        "spyder/dark/string": ("#11a642", False, True),
        "spyder/dark/number": ("#c80000", False, False),
        "spyder/dark/instance": ("#be5f00", False, True),
        # ---- Zenburn ----
        "zenburn/name": "Zenburn",
        #        Name            Color     Bold  Italic
        "zenburn/background": "#1f1f1f",
        "zenburn/currentline": "#333333",
        "zenburn/currentcell": "#2c2c2c",
        "zenburn/occurrence": "#7a738f",
        "zenburn/ctrlclick": "#0000ff",
        "zenburn/sideareas": "#3f3f3f",
        "zenburn/matched_p": "#688060",
        "zenburn/unmatched_p": "#bd6e76",
        "zenburn/normal": ("#dcdccc", False, False),
        "zenburn/keyword": ("#dfaf8f", True, False),
        "zenburn/builtin": ("#efef8f", False, False),
        "zenburn/definition": ("#efef8f", False, False),
        "zenburn/comment": ("#7f9f7f", False, True),
        "zenburn/string": ("#cc9393", False, False),
        "zenburn/number": ("#8cd0d3", False, False),
        "zenburn/instance": ("#dcdccc", False, True),
        # ---- Solarized Light ----
        "solarized/light/name": "Solarized Light",
        #        Name            Color     Bold  Italic
        "solarized/light/background": "#fdf6e3",
        "solarized/light/currentline": "#f5efdB",
        "solarized/light/currentcell": "#eee8d5",
        "solarized/light/occurrence": "#839496",
        "solarized/light/ctrlclick": "#d33682",
        "solarized/light/sideareas": "#eee8d5",
        "solarized/light/matched_p": "#586e75",
        "solarized/light/unmatched_p": "#dc322f",
        "solarized/light/normal": ("#657b83", False, False),
        "solarized/light/keyword": ("#859900", False, False),
        "solarized/light/builtin": ("#6c71c4", False, False),
        "solarized/light/definition": ("#268bd2", True, False),
        "solarized/light/comment": ("#93a1a1", False, True),
        "solarized/light/string": ("#2aa198", False, False),
        "solarized/light/number": ("#cb4b16", False, False),
        "solarized/light/instance": ("#b58900", False, True),
        # ---- Solarized Dark ----
        "solarized/dark/name": "Solarized Dark",
        #        Name            Color     Bold  Italic
        "solarized/dark/background": "#1f1f1f",
        "solarized/dark/currentline": "#083f4d",
        "solarized/dark/currentcell": "#073642",
        "solarized/dark/occurrence": "#657b83",
        "solarized/dark/ctrlclick": "#d33682",
        "solarized/dark/sideareas": "#073642",
        "solarized/dark/matched_p": "#93a1a1",
        "solarized/dark/unmatched_p": "#dc322f",
        "solarized/dark/normal": ("#839496", False, False),
        "solarized/dark/keyword": ("#859900", False, False),
        "solarized/dark/builtin": ("#6c71c4", False, False),
        "solarized/dark/definition": ("#268bd2", True, False),
        "solarized/dark/comment": ("#586e75", False, True),
        "solarized/dark/string": ("#2aa198", False, False),
        "solarized/dark/number": ("#cb4b16", False, False),
        "solarized/dark/instance": ("#b58900", False, True),
    },
}

CONF = UserConfig(DEFAULTS)


class ValidationMode(enum.Enum):
    """Type checking modes for DataItems"""

    DISABLED = 0  # No checking at all
    ENABLED = 1  # Warnings on validation error
    STRICT = 2  # Exceptions on validation error


# Internal (non-user) runtime configuration
class InternalConfig:
    """Internal guidata configuration flags"""

    def __init__(self):
        # Validation mode is disabled by default for backward compatibility
        self.validation_mode = ValidationMode.DISABLED


_INTERNAL_CONF = InternalConfig()


def set_validation_mode(mode: ValidationMode) -> None:
    """Set the validation mode for DataItems"""
    if not isinstance(mode, ValidationMode):
        raise ValueError("Invalid validation mode")
    _INTERNAL_CONF.validation_mode = mode


def get_validation_mode() -> ValidationMode:
    """Get the current validation mode for DataItems"""
    return _INTERNAL_CONF.validation_mode


# Add a context manager to temporarily set validation mode
@contextlib.contextmanager
def temporary_validation_mode(mode: ValidationMode) -> Generator[None, None, None]:
    """Temporarily set the validation mode for DataItems"""
    original_mode = get_validation_mode()
    set_validation_mode(mode)
    try:
        yield
    finally:
        set_validation_mode(original_mode)



# -*- coding: utf-8 -*-
#
# Copyright © Spyder Project Contributors
# Licensed under the terms of the MIT License


"""
Tests for arrayeditor.py
"""

# guitest: show

import numpy as np

from guidata.env import execenv
from guidata.qthelpers import exec_dialog, qt_app_context
from guidata.widgets.arrayeditor import ArrayEditor


def launch_arrayeditor(data, title="", xlabels=None, ylabels=None, variable_size=False):
    """Helper routine to launch an arrayeditor and return its result"""
    dlg = ArrayEditor()
    dlg.setup_and_check(
        data, title, xlabels=xlabels, ylabels=ylabels, variable_size=variable_size
    )
    exec_dialog(dlg)
    return dlg.get_value()


def test_arrayeditor():
    """Test array editor for all supported data types"""
    with qt_app_context():
        # Variable size version
        for title, data in (
            ("string array", np.array(["kjrekrjkejr"])),
            (
                "masked array",
                np.ma.array([[1, 0], [1, 0]], mask=[[True, False], [False, False]]),
            ),
            ("int array", np.array([1, 2, 3], dtype="int8")),
        ):
            launch_arrayeditor(data, "[Variable size] " + title, variable_size=True)
        for title, data in (
            ("string array", np.array(["kjrekrjkejr"])),
            ("unicode array", np.array(["ñññéáíó"])),
            (
                "masked array",
                np.ma.array([[1, 0], [1, 0]], mask=[[True, False], [False, False]]),
            ),
            (
                "record array",
                np.zeros(
                    (2, 2),
                    {
                        "names": ("red", "green", "blue"),
                        "formats": (np.float32, np.float32, np.float32),
                    },
                ),
            ),
            (
                "record array with titles",
                np.array(
                    [(0, 0.0), (0, 0.0), (0, 0.0)],
                    dtype=[(("title 1", "x"), "|i1"), (("title 2", "y"), ">f4")],
                ),
            ),
            ("bool array", np.array([True, False, True])),
            ("int array", np.array([1, 2, 3], dtype="int8")),
            ("float16 array", np.zeros((5, 5), dtype=np.float16)),
        ):
            launch_arrayeditor(data, title)
        for title, data, xlabels, ylabels in (
            ("float array", np.random.rand(5, 5), ["a", "b", "c", "d", "e"], None),
            (
                "complex array",
                np.round(np.random.rand(5, 5) * 10)
                + np.round(np.random.rand(5, 5) * 10) * 1j,
                np.linspace(-12, 12, 5),
                np.linspace(-12, 12, 5),
            ),
        ):
            launch_arrayeditor(data, title, xlabels, ylabels)

        arr = np.zeros((3, 3, 4))
        arr[0, 0, 0] = 1
        arr[0, 0, 1] = 2
        arr[0, 0, 2] = 3
        launch_arrayeditor(arr, "3D array")

        execenv.print("OK")


if __name__ == "__main__":
    test_arrayeditor()



#!/usr/bin/env python3
"""
Test script for SeparatorItem trailing separator filtering functionality
"""

import pytest

from guidata.dataset import DataSet, IntItem, SeparatorItem, StringItem


class DataSetNormal(DataSet):
    """Dataset with separator not at the end"""

    name = StringItem("Name", default="test")
    separator1 = SeparatorItem("sep1")
    value = IntItem("Value", default=42)
    separator2 = SeparatorItem("sep2")
    another_value = IntItem("Another Value", default=100)


class DataSetWithTrailingSeparator(DataSet):
    """Dataset with trailing separators"""

    name = StringItem("Name", default="test")
    separator1 = SeparatorItem("sep1")
    value = IntItem("Value", default=42)
    separator2 = SeparatorItem("sep2")
    separator3 = SeparatorItem("sep3")  # Trailing separator


class DataSetOnlySeparators(DataSet):
    """Dataset with only separators"""

    separator1 = SeparatorItem("sep1")
    separator2 = SeparatorItem("sep2")
    separator3 = SeparatorItem("sep3")


class DataSetMultipleTrailingSeparators(DataSet):
    """Dataset with multiple trailing separators"""

    name = StringItem("Name", default="test")
    separator1 = SeparatorItem("sep1")
    separator2 = SeparatorItem("sep2")
    separator3 = SeparatorItem("sep3")
    separator4 = SeparatorItem("sep4")


class DataSetEmpty(DataSet):
    """Empty dataset"""

    pass


def test_normal_dataset_no_trailing_separator():
    """Test dataset with separator not at the end - no filtering should occur"""
    ds = DataSetNormal()
    items = ds.get_items()
    item_names = [item._name for item in items]

    # Should have all items since separator is not trailing
    assert len(items) == 5
    assert item_names == ["name", "separator1", "value", "separator2", "another_value"]

    # Check string representation
    string_repr = str(ds)
    assert "sep1: -" in string_repr
    assert "sep2: -" in string_repr
    assert "Another Value: 100" in string_repr


def test_trailing_separator_filtering():
    """Test dataset with trailing separators - they should be filtered out"""
    ds = DataSetWithTrailingSeparator()
    items = ds.get_items()
    item_names = [item._name for item in items]

    # Should have 5 items (trailing separators NOT filtered in get_items())
    assert len(items) == 5
    assert item_names == ["name", "separator1", "value", "separator2", "separator3"]
    assert items[-1]._name == "separator3"  # get_items() includes trailing separators

    # Check string representation doesn't include trailing separators
    string_repr = str(ds)
    assert "sep1: -" in string_repr
    assert "sep2: -" not in string_repr  # Trailing separator should not appear
    assert "sep3: -" not in string_repr  # Trailing separator should not appear


def test_only_separators_dataset():
    """Test dataset with only separators - all should be filtered out"""
    ds = DataSetOnlySeparators()
    items = ds.get_items()

    # Should have 3 items (separators NOT filtered in get_items())
    assert len(items) == 3

    # Check string representation
    string_repr = str(ds)
    lines = string_repr.strip().split("\n")
    # Should only contain the dataset title
    assert len(lines) == 1
    assert "Dataset with only separators:" in lines[0]


def test_multiple_trailing_separators():
    """Test dataset with multiple trailing separators - all should be filtered"""
    ds = DataSetMultipleTrailingSeparators()
    items = ds.get_items()
    item_names = [item._name for item in items]

    # Should have 5 items (trailing separators NOT filtered in get_items())
    assert len(items) == 5
    expected_names = ["name", "separator1", "separator2", "separator3", "separator4"]
    assert item_names == expected_names

    # Check string representation
    string_repr = str(ds)
    assert "sep1: -" not in string_repr  # All separators should be filtered
    assert "sep2: -" not in string_repr
    assert "sep3: -" not in string_repr
    assert "sep4: -" not in string_repr


def test_empty_dataset():
    """Test empty dataset"""
    ds = DataSetEmpty()
    items = ds.get_items()

    # Should have 0 items
    assert len(items) == 0

    # Check string representation
    string_repr = str(ds)
    lines = string_repr.strip().split("\n")
    assert len(lines) == 1
    assert "Empty dataset:" in lines[0]


def test_get_items_copy_behavior():
    """Test that the copy parameter works correctly with filtering"""
    ds = DataSetWithTrailingSeparator()

    # Test without copy
    items1 = ds.get_items(copy=False)
    items2 = ds.get_items(copy=False)

    # Both should have 5 items (no filtering in get_items())
    assert len(items1) == 5
    assert len(items2) == 5

    # Test with copy
    items3 = ds.get_items(copy=True)
    items4 = ds.get_items(copy=True)

    # Both should have 5 items (no filtering in get_items())
    assert len(items3) == 5
    assert len(items4) == 5  # Copied items should be different objects
    assert items3 is not items4
    assert items3[0] is not items4[0]


def test_string_representation_consistency():
    """Test that string representation is consistent with get_items filtering"""
    ds = DataSetWithTrailingSeparator()
    items = ds.get_items()
    string_repr = str(ds)

    # Count the number of items in string representation
    lines = [line.strip() for line in string_repr.split("\n") if ":" in line]
    # Remove the dataset title line
    data_lines = [line for line in lines if not line.endswith(":")]

    # String representation should have fewer lines due to trailing separator filtering
    # get_items() returns 5 items, but string representation should show only 3 lines
    assert len(items) == 5  # get_items includes trailing separators
    assert len(data_lines) == 3  # but string representation filters them


def test_gui_trailing_separator_filtering():
    """Test that GUI widgets filter trailing separators correctly"""
    try:
        from guidata.dataset.qtwidgets import DataSetEditGroupBox, DataSetShowGroupBox
        from guidata.qthelpers import qt_app_context

        with qt_app_context():
            # Test the edit widget
            dataset = DataSetWithTrailingSeparator()

            edit_group = DataSetEditGroupBox("Edit Test", DataSetWithTrailingSeparator)
            edit_group.instance = dataset

            show_group = DataSetShowGroupBox("Show Test", DataSetWithTrailingSeparator)
            show_group.instance = dataset

            # Count the number of child widgets that are actual item widgets
            edit_widgets = [w for w in edit_group.edit.widgets if hasattr(w, "item")]
            show_widgets = [w for w in show_group.edit.widgets if hasattr(w, "item")]

            # Should have 3 widgets (name, separator1, value)
            #  - trailing separators filtered
            expected_count = 3
            assert len(edit_widgets) == expected_count
            assert len(show_widgets) == expected_count

    except ImportError:
        # Skip GUI tests if Qt is not available
        pytest.skip("Qt not available for GUI testing")


if __name__ == "__main__":
    pytest.main([__file__])
